(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var _ = Package.underscore._;
var Random = Package.random.Random;
var Router = Package['iron:router'].Router;
var RouteController = Package['iron:router'].RouteController;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var check = Package.check.check;
var Match = Package.check.Match;
var SHA256 = Package.sha.SHA256;
var Email = Package.email.Email;
var EmailInternals = Package.email.EmailInternals;
var Iron = Package['iron:core'].Iron;

(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                   //
// packages/accounts-email-token/token_common.js                                                     //
//                                                                                                   //
///////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                     //
Accounts.emailToken = {};                                                                            // 1
                                                                                                     // 2
Accounts.emailToken._hashToken = function (token) {                                                  // 3
  return {                                                                                           // 4
    digest: SHA256(token),                                                                           // 5
    algorithm: "sha-256"                                                                             // 6
  };                                                                                                 // 7
};                                                                                                   // 8
                                                                                                     // 9
var enabled = false;                                                                                 // 10
                                                                                                     // 11
if (Meteor.isClient) {                                                                               // 12
  var enabledReactive = new ReactiveVar(false);                                                      // 13
}                                                                                                    // 14
                                                                                                     // 15
Accounts.emailToken.enable = function (accountsUi) {                                                 // 16
  enabled = true;                                                                                    // 17
  if (Meteor.isClient) {                                                                             // 18
    accountsUi.registerService("emailToken", "an Email + Token");                                    // 19
    enabledReactive.set(true);                                                                       // 20
  }                                                                                                  // 21
};                                                                                                   // 22
                                                                                                     // 23
Accounts.emailToken.disable = function (accountsUi) {                                                // 24
  enabled = false;                                                                                   // 25
  if (Meteor.isClient) {                                                                             // 26
    accountsUi.deregisterService("emailToken");                                                      // 27
    enabledReactive.set(false);                                                                      // 28
  }                                                                                                  // 29
};                                                                                                   // 30
                                                                                                     // 31
Accounts.emailToken.isEnabled = function () {                                                        // 32
  if (Meteor.isClient) {                                                                             // 33
    return enabledReactive.get();                                                                    // 34
  } else {                                                                                           // 35
    return enabled;                                                                                  // 36
  }                                                                                                  // 37
};                                                                                                   // 38
                                                                                                     // 39
Router.route("/_emailLogin/:_email/:_token", function () {                                           // 40
  this.render("Loading");                                                                            // 41
  var that = this;                                                                                   // 42
  Meteor.loginWithEmailToken(this.params._email, this.params._token, function (err) {                // 43
    if (err) {                                                                                       // 44
      that.render("_emailTokenError", {                                                              // 45
        data: function () {                                                                          // 46
          return {                                                                                   // 47
            error: err                                                                               // 48
          };                                                                                         // 49
        }                                                                                            // 50
      });                                                                                            // 51
    } else {                                                                                         // 52
      Router.go("/");                                                                                // 53
    }                                                                                                // 54
  });                                                                                                // 55
});                                                                                                  // 56
                                                                                                     // 57
Router.route("/_emailLinkIdentity/:_email/:_token/:_accountId", function () {                        // 58
  this.render("Loading");                                                                            // 59
  if (this.state.get("error")) {                                                                     // 60
    this.render("_emailLinkIdentityError", {data: {error: this.state.get("error")}});                // 61
  } else {                                                                                           // 62
    var self = this;                                                                                 // 63
    if (Meteor.userId() === this.params._accountId) {                                                // 64
      Meteor.call("linkEmailIdentityToAccount",                                                      // 65
                  this.params._email, this.params._token, function (err, result) {                   // 66
        if (err) {                                                                                   // 67
          self.state.set("error", err.toString());                                                   // 68
        } else {                                                                                     // 69
          Router.go("/account");                                                                     // 70
        }                                                                                            // 71
      });                                                                                            // 72
    } else {                                                                                         // 73
      this.render("_emailLinkIdentityError");                                                        // 74
    }                                                                                                // 75
  }                                                                                                  // 76
});                                                                                                  // 77
                                                                                                     // 78
///////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                   //
// packages/accounts-email-token/token_server.js                                                     //
//                                                                                                   //
///////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                     //
var TOKEN_EXPIRATION_MS = 15 * 60 * 1000;                                                            // 1
                                                                                                     // 2
var cleanupExpiredTokens = function() {                                                              // 3
  Meteor.users.update(                                                                               // 4
      {"services.email.tokens.createdAt": {$lt: new Date(Date.now() - TOKEN_EXPIRATION_MS)}},        // 5
      {$pull: {                                                                                      // 6
        "services.email.tokens": {                                                                   // 7
          createdAt: {$lt: new Date(Date.now() - TOKEN_EXPIRATION_MS)}}}},                           // 8
    { multi: true });                                                                                // 9
};                                                                                                   // 10
                                                                                                     // 11
Meteor.startup(cleanupExpiredTokens);                                                                // 12
// Tokens can actually last up to 2 * TOKEN_EXPIRATION_MS                                            // 13
SandstormDb.periodicCleanup(TOKEN_EXPIRATION_MS, cleanupExpiredTokens);                              // 14
                                                                                                     // 15
var checkToken = function (tokens, token) {                                                          // 16
  var found = false;                                                                                 // 17
  tokens.forEach(function (userToken) {                                                              // 18
    if((userToken.algorithm === token.algorithm) &&                                                  // 19
       (userToken.digest === token.digest)) {                                                        // 20
      found = true;                                                                                  // 21
    }                                                                                                // 22
  });                                                                                                // 23
                                                                                                     // 24
  return found;                                                                                      // 25
};                                                                                                   // 26
                                                                                                     // 27
// The name of the email package to use. It refers to a variable named in the global scope.          // 28
var EMAIL_PACKAGE= "Email";                                                                          // 29
                                                                                                     // 30
Accounts.emailToken.setEmailPackage = function (packageName) {                                       // 31
  EMAIL_PACKAGE = packageName;                                                                       // 32
};                                                                                                   // 33
                                                                                                     // 34
function consumeToken (user, token) {                                                                // 35
  var hashedToken = Accounts.emailToken._hashToken(token);                                           // 36
  var found = checkToken(user.services.email.tokens, hashedToken);                                   // 37
                                                                                                     // 38
  if (found) {                                                                                       // 39
    Meteor.users.update({_id: user._id}, {$pull: {"services.email.tokens": hashedToken}});           // 40
  }                                                                                                  // 41
                                                                                                     // 42
  return found;                                                                                      // 43
}                                                                                                    // 44
                                                                                                     // 45
// Handler to login with a token.                                                                    // 46
Accounts.registerLoginHandler("email", function (options) {                                          // 47
  if (!options.email)                                                                                // 48
    return undefined; // don't handle                                                                // 49
                                                                                                     // 50
  options = options.email;                                                                           // 51
  check(options, {                                                                                   // 52
    email: String,                                                                                   // 53
    token: String                                                                                    // 54
  });                                                                                                // 55
                                                                                                     // 56
  var user = Meteor.users.findOne({"services.email.email": options.email},                           // 57
                                  {fields: {"services.email": 1}});                                  // 58
                                                                                                     // 59
  if (!user) {                                                                                       // 60
    console.error("User not found:", options.email);                                                 // 61
    return {                                                                                         // 62
      error: new Meteor.Error(403, "User not found")                                                 // 63
    };                                                                                               // 64
  }                                                                                                  // 65
                                                                                                     // 66
  if (!user.services.email.tokens) {                                                                 // 67
    console.error("User has no token set:", options.email);                                          // 68
    return {                                                                                         // 69
      error: new Meteor.Error(403, "User has no token set")                                          // 70
    };                                                                                               // 71
  }                                                                                                  // 72
                                                                                                     // 73
  if (!consumeToken(user, options.token.trim())) {                                                   // 74
    console.error("Token not found:", options.email);                                                // 75
    return {                                                                                         // 76
      error: new Meteor.Error(403, "Invalid authentication code")                                    // 77
    };                                                                                               // 78
  }                                                                                                  // 79
                                                                                                     // 80
  return {                                                                                           // 81
    userId: user._id                                                                                 // 82
  };                                                                                                 // 83
});                                                                                                  // 84
                                                                                                     // 85
var Url = Npm.require("url");                                                                        // 86
                                                                                                     // 87
var ROOT_URL = Url.parse(process.env.ROOT_URL);                                                      // 88
var HOSTNAME = ROOT_URL.hostname;                                                                    // 89
                                                                                                     // 90
var makeTokenUrl = function (email, token, linkingIdentity) {                                        // 91
  if (linkingIdentity) {                                                                             // 92
    return process.env.ROOT_URL + "/_emailLinkIdentity/" + encodeURIComponent(email) + "/" +         // 93
        encodeURIComponent(token) + "/" + Meteor.userId();                                           // 94
  } else {                                                                                           // 95
    return process.env.ROOT_URL + "/_emailLogin/" + encodeURIComponent(email) + "/" +                // 96
        encodeURIComponent(token);                                                                   // 97
  }                                                                                                  // 98
};                                                                                                   // 99
                                                                                                     // 100
///                                                                                                  // 101
/// EMAIL VERIFICATION                                                                               // 102
///                                                                                                  // 103
var sendTokenEmail = function (email, token, linkingIdentity) {                                      // 104
  var subject;                                                                                       // 105
  var text;                                                                                          // 106
  if (!linkingIdentity) {                                                                            // 107
    subject = "Log in to " + HOSTNAME;                                                               // 108
    text = "To confirm this email address on ";                                                      // 109
  } else {                                                                                           // 110
    subject = "Confirm this email address on " + HOSTNAME;                                           // 111
    text = "To confirm this email address on ";                                                      // 112
  }                                                                                                  // 113
  text = text + HOSTNAME + ", click on the following link:\n\n" +                                    // 114
      makeTokenUrl(email, token, linkingIdentity) + "\n\n" +                                         // 115
      "Alternatively, enter the following one-time authentication code into the log-in form:\n\n" +  // 116
      token;                                                                                         // 117
                                                                                                     // 118
  var options = {                                                                                    // 119
    to:  email,                                                                                      // 120
    from: HOSTNAME + " <no-reply@" + HOSTNAME + ">",                                                 // 121
    subject: subject,                                                                                // 122
    text: text,                                                                                      // 123
  };                                                                                                 // 124
                                                                                                     // 125
  global[EMAIL_PACKAGE].send(options);                                                               // 126
};                                                                                                   // 127
                                                                                                     // 128
///                                                                                                  // 129
/// CREATING USERS                                                                                   // 130
///                                                                                                  // 131
// returns the user id                                                                               // 132
var createAndEmailTokenForUser = function (email, linkingIdentity) {                                 // 133
  check(email, String);                                                                              // 134
  check(linkingIdentity, Boolean);                                                                   // 135
  var atIndex = email.indexOf("@");                                                                  // 136
  if (atIndex === -1) {                                                                              // 137
    throw new Meteor.Error(400, "No @ symbol was found in your email");                              // 138
  }                                                                                                  // 139
                                                                                                     // 140
  var user = Meteor.users.findOne({"services.email.email": email},                                   // 141
                                  {fields: {"services.email": 1}});                                  // 142
  var userId;                                                                                        // 143
                                                                                                     // 144
  // TODO(someday): make this shorter, and handle requests that try to brute force it.               // 145
  var token = Random.id(12);                                                                         // 146
  var tokenObj = Accounts.emailToken._hashToken(token);                                              // 147
  tokenObj.createdAt = new Date();                                                                   // 148
                                                                                                     // 149
  if (user) {                                                                                        // 150
    if (user.services.email.tokens && user.services.email.tokens.length > 2) {                       // 151
      throw new Meteor.Error(409, "It looks like we sent a log in email to this address not long " +
        "ago. Please use the one that was already sent (check your spam folder if you can't find " +
        "it), or wait a while and try again");                                                       // 154
    }                                                                                                // 155
    userId = user._id;                                                                               // 156
                                                                                                     // 157
    Meteor.users.update({_id: user._id}, {$push: {"services.email.tokens": tokenObj}});              // 158
  } else {                                                                                           // 159
    var options = {};                                                                                // 160
    user = {services: {email: {                                                                      // 161
      tokens: [tokenObj],                                                                            // 162
      email: email                                                                                   // 163
    }}};                                                                                             // 164
                                                                                                     // 165
    userId = Accounts.insertUserDoc(options, user);                                                  // 166
  }                                                                                                  // 167
                                                                                                     // 168
  sendTokenEmail(email, token, linkingIdentity);                                                     // 169
                                                                                                     // 170
  return userId;                                                                                     // 171
};                                                                                                   // 172
                                                                                                     // 173
Meteor.methods({                                                                                     // 174
  createAndEmailTokenForUser: function (email, linkingIdentity) {                                    // 175
    // method for create user. Requests come from the client.                                        // 176
    // This method will create a user if it doesn't exist, otherwise it will generate a token.       // 177
    // It will always send an email to the user                                                      // 178
                                                                                                     // 179
    check(email, String);                                                                            // 180
    check(linkingIdentity, Boolean);                                                                 // 181
                                                                                                     // 182
    if (!Accounts.emailToken.isEnabled()) {                                                          // 183
      throw new Meteor.Error(403, "Email identity service is disabled.");                            // 184
    }                                                                                                // 185
    // Create user. result contains id and token.                                                    // 186
    var user = createAndEmailTokenForUser(email, linkingIdentity);                                   // 187
  },                                                                                                 // 188
                                                                                                     // 189
  linkEmailIdentityToAccount: function (email, token) {                                              // 190
    // Links the email identity with address `email` and login token `token` to the current account.
    check(email, String);                                                                            // 192
    check(token, String);                                                                            // 193
    var account = Meteor.user();                                                                     // 194
    if (!account || !account.loginIdentities) {                                                      // 195
      throw new Meteor.Error(403, "Must be logged in to an account to link an email identity.");     // 196
    }                                                                                                // 197
    var identity = Meteor.users.findOne({"services.email.email": email},                             // 198
                                        {fields: {"services.email": 1}});                            // 199
    if (!identity) {                                                                                 // 200
      throw new Meteor.Error(403, "Invalid authentication code.");                                   // 201
    }                                                                                                // 202
    if (!consumeToken(identity, token)) {                                                            // 203
      throw new Meteor.Error(403, "Invalid authentication code.");                                   // 204
    }                                                                                                // 205
    Accounts.linkIdentityToAccount(identity._id, account._id);                                       // 206
  }                                                                                                  // 207
});                                                                                                  // 208
                                                                                                     // 209
Meteor.users._ensureIndex("services.email.email", {unique: 1, sparse: 1});                           // 210
                                                                                                     // 211
///////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['accounts-email-token'] = {};

})();

//# sourceMappingURL=accounts-email-token.js.map
