(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;

/* Package-scope variables */
var Packages, DevPackages, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, AppIndex, KeybaseProfiles, isDemoUser, isSignedUp, isSignedUpOrDemo, successfulReferralsCount, getUserQuota, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, roleAssignmentPattern, SandstormDb, i, appName, addStaticAsset, verifiedEmails, unverifiedEmails, userPictureUrl, fetchPicture;

(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/db.js                                                                               //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
// This file defines the database schema.                                                                    // 17
                                                                                                             // 18
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every               // 19
// query it makes, so you can see if it's doing queries too often, etc.                                      // 20
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                                      // 21
  var oldFind = Mongo.Collection.prototype.find;                                                             // 22
  Mongo.Collection.prototype.find = function () {                                                            // 23
    console.log(this._prefix, arguments);                                                                    // 24
    return oldFind.apply(this, arguments);                                                                   // 25
  }                                                                                                          // 26
}                                                                                                            // 27
                                                                                                             // 28
// Users = new Mongo.Collection("users");                                                                    // 29
// The users collection is special and can be accessed through `Meteor.users`.                               // 30
// See https://docs.meteor.com/#/full/meteor_users.                                                          // 31
//                                                                                                           // 32
// There are two distinct types of entries in the users collection: identities and accounts. An              // 33
// identity contains personal profile information and typically includes some intrinsic method for           // 34
// authenticating as the owner of that information.                                                          // 35
//                                                                                                           // 36
// An account is an owner of app actions, grains, contacts, notifications, and payment info.                 // 37
// Each account can have multiple identities linked to it. To log in as an account you must first            // 38
// authenticate as one of its linked identities.                                                             // 39
//                                                                                                           // 40
// Every user contains the following fields:                                                                 // 41
//   _id: Unique string ID. For accounts, this is random. For identities, this is the globally               // 42
//        stable SHA-256 ID of this identity, hex-encoded.                                                   // 43
//   createdAt: Date when this entry was added to the collection.                                            // 44
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.                      // 45
//   services: Object containing login data used by Meteor authentication services.                          // 46
//   expires: Date when this user should be deleted. Only present for demo users.                            // 47
//   upgradedFromDemo: If present, the date when this user was upgraded from being a demo user.              // 48
//                                                                                                           // 49
// Identity users additionally contain the following fields:                                                 // 50
//   profile: Object containing the data that will be shared with users and grains that come into            // 51
//            contact with this identity. Includes the following fields:                                     // 52
//       service: String containing the name of this identity's authentication method.                       // 53
//       name: String containing the chosen display name of the identity.                                    // 54
//       handle: String containing the identity's preferred handle.                                          // 55
//       picture: _id into the StaticAssets table for the identity's picture. If not present,                // 56
//                an identicon will be used.                                                                 // 57
//       pronoun: One of "male", "female", "neutral", or "robot".                                            // 58
//   unverifiedEmail: If present, a string containing an email address specified by the user.                // 59
//   referredBy: ID of the Account that referred this Identity.                                              // 60
//                                                                                                           // 61
// Account users additionally contain the following fields:                                                  // 62
//   loginIdentities: Array of identity objects, each of which may include the following fields.             // 63
//       id: The globally-stable SHA-256 ID of this identity, hex-encoded.                                   // 64
//   nonloginIdentities: Array of identity objects, of the same form as `loginIdentities`. We use            // 65
//                       a separate array here so that we can use a Mongo index to enforce the               // 66
//                       invariant that an identity only be a login identity for a single account.           // 67
//   primaryEmail: String containing this account's primary email address. Must be a verified adress         // 68
//                 of one of this account's linked identities. Call SandstormDb.getUserEmails()              // 69
//                 to do this checking automatically.                                                        // 70
//   isAdmin: Boolean indicating whether this account is allowed to access the Sandstorm admin panel.        // 71
//   signupKey: If this is an invited user, then this field contains their signup key.                       // 72
//   signupNote: If the user was invited through a link, then this field contains the note that the          // 73
//               inviter admin attached to the key.                                                          // 74
//   signupEmail: If the user was invited by email, then this field contains the email address that          // 75
//                the invite was sent to.                                                                    // 76
//   hasCompletedSignup: True if this account has confirmed its profile and agreed to this server's          // 77
//                       terms of service.                                                                   // 78
//   plan: _id of an entry in the Plans table which determines the user's qutoa.                             // 79
//   storageUsage: Number of bytes this user is currently storing.                                           // 80
//   isAppDemoUser: True if this is a demo user who arrived via an /appdemo/ link.                           // 81
//   appDemoId: If this is an appdemo user (see above), the app ID they started out demoing.                 // 82
//   payments: Object defined by payments module, if loaded.                                                 // 83
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                       // 84
//   referredByComplete: ID of the Account that referred this Account. If this is set, we                    // 85
//                        stop writing new referredBy values onto Identities for this account.               // 86
//   referredCompleteDate: The Date at which the completed referral occurred.                                // 87
//   referredIdentityIds: List of Identity IDs that this Account has referred. This is used for              // 88
//                        reliably determining which Identity's names are safe to display.                   // 89
//   stashedOldUser: A complete copy of this user from before the accounts/identities migration.             // 90
//                   TODO(cleanup): Delete this field once we're sure it's safe to do so.                    // 91
                                                                                                             // 92
Packages = new Mongo.Collection("packages");                                                                 // 93
// Packages which are installed or downloading.                                                              // 94
//                                                                                                           // 95
// Each contains:                                                                                            // 96
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                          // 97
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"         // 98
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),                      // 99
//       >1 = download byte count.                                                                           // 100
//   error:  If status is "failed", error message string.                                                    // 101
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.                // 102
//   appId:  If status is "ready", the application ID string.  Packages representing different               // 103
//       versions of the same app have the same appId.  The spk tool defines the app ID format               // 104
//       and can cryptographically verify that a package belongs to a particular app ID.                     // 105
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package              // 106
//       collector should at some point check whether there are any other references and, if not,            // 107
//       delete the package.                                                                                 // 108
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.               // 109
//   isAutoUpdated: This package was downloaded as part of an auto-update. We shouldn't clean it up          // 110
//     even if it has no users.                                                                              // 111
//   authorPgpKeyFingerprint: Verified PGP key fingerprint (SHA-1, hex, all-caps) of the app                 // 112
//     packager.                                                                                             // 113
                                                                                                             // 114
DevPackages = new Mongo.Collection("devpackages");                                                           // 115
// List of packages currently made available via the dev tools running on the local machine.                 // 116
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool           // 117
// on the local machine to publish an under-development app to this server. That should only ever            // 118
// happen on developers' desktop machines.                                                                   // 119
//                                                                                                           // 120
// While a dev package is published, it automatically appears as installed by every user of the              // 121
// server, and it overrides all packages with the same application ID. If any instances of those             // 122
// packages are currently open, they are killed and reset on publish.                                        // 123
//                                                                                                           // 124
// When the dev tool disconnects, the package is automatically unpublished, and any open instances           // 125
// are again killed and refreshed.                                                                           // 126
//                                                                                                           // 127
// Each contains:                                                                                            // 128
//   _id:  The package ID string (as with Packages._id).                                                     // 129
//   appId: The app ID this package is intended to override (as with Packages.appId).                        // 130
//   timestamp:  Time when the package was last updated. If this changes while the package is                // 131
//     published, all running instances are reset. This is used e.g. to reset the app each time              // 132
//     changes are made to the source code.                                                                  // 133
//   manifest:  The app's manifest, as with Packages.manifest.                                               // 134
                                                                                                             // 135
UserActions = new Mongo.Collection("userActions");                                                           // 136
// List of actions that each user has installed which create new grains.  Each app may install               // 137
// some number of actions (usually, one).                                                                    // 138
//                                                                                                           // 139
// Each contains:                                                                                            // 140
//   _id:  random                                                                                            // 141
//   userId:  User who has installed this action.                                                            // 142
//   packageId:  Package used to run this action.                                                            // 143
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 144
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle; denormalized so                       // 145
//       that clients can access it without subscribing to the Packages collection.                          // 146
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 147
//       searchability.                                                                                      // 148
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"                     // 149
//   title: JSON-encoded LocalizedText title for this action, e.g.                                           // 150
//       `{defaultText: "New Spreadsheet"}`.                                                                 // 151
//   nounPhrase: JSON-encoded LocalizedText describing what is created when this action is run.              // 152
//   command:  Manifest.Command to run this action (see package.capnp).                                      // 153
                                                                                                             // 154
Grains = new Mongo.Collection("grains");                                                                     // 155
// Grains belonging to users.                                                                                // 156
//                                                                                                           // 157
// Each contains:                                                                                            // 158
//   _id:  random                                                                                            // 159
//   packageId:  _id of the package of which this grain is an instance.                                      // 160
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                      // 161
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                  // 162
//       searchability.                                                                                      // 163
//   userId: The _id of the account that owns this grain.                                                    // 164
//   identityId: The identity with which the owning account prefers to open this grain.                      // 165
//   title:  Human-readable string title, as chosen by the user.                                             // 166
//   lastUsed:  Date when the grain was last used by a user.                                                 // 167
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                          // 168
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent          // 169
//                   time a session to this grain was opened.                                                // 170
//                                                                                                           // 171
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and           // 172
// web publishing functionality without powerbox support; they will be replaced once the powerbox            // 173
// is implemented.                                                                                           // 174
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and           // 175
//       web publishing. This field is initialized when first requested by the app.                          // 176
                                                                                                             // 177
RoleAssignments = new Mongo.Collection("roleAssignments");                                                   // 178
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges          // 179
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens                   // 180
// collection.                                                                                               // 181
                                                                                                             // 182
Contacts = new Mongo.Collection("contacts");                                                                 // 183
// Edges in the social graph.                                                                                // 184
//                                                                                                           // 185
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob               // 186
// will show up in her user-picker UI for actions like share-by-identity.                                    // 187
//                                                                                                           // 188
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of              // 189
// Bob's.                                                                                                    // 190
//                                                                                                           // 191
// Each contains:                                                                                            // 192
//   _id: random                                                                                             // 193
//   ownerId: The `_id` of the user who owns this contact.                                                   // 194
//   identityId:  The identity of the contacted user.                                                        // 195
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies              // 196
//            the contact to the owner.                                                                      // 197
//   created: Date when this contact was created.                                                            // 198
                                                                                                             // 199
Sessions = new Mongo.Collection("sessions");                                                                 // 200
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.          // 201
//                                                                                                           // 202
// Each contains:                                                                                            // 203
//   _id:  random                                                                                            // 204
//   grainId:  _id of the grain to which this session is connected.                                          // 205
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the           // 206
//       '*' in WILDCARD_HOST.                                                                               // 207
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some              // 208
//       period.                                                                                             // 209
//   userId:  User ID of the user who owns this session.                                                     // 210
//   identityId:  Identity ID of the user who owns this session.                                             // 211
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens            // 212
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),               // 213
//       anonymous users can get access without an API token and so neither userId nor hashedToken           // 214
//       are present.                                                                                        // 215
//   powerboxView: If present, this is a view that should be presented as part of a powerbox                 // 216
//       interaction.                                                                                        // 217
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                         // 218
//   viewInfo: The UiView.ViewInfo corresponding to the underlying UiSession. This isn't populated           // 219
//       until newSession is called on the UiView.                                                           // 220
//   permissions: The permissions for the current identity on this UiView. This isn't populated              // 221
//       until newSession is called on the UiView.                                                           // 222
//   hasLoaded: Marked as true by the proxy when the underlying UiSession has responded to its first         // 223
//       request                                                                                             // 224
                                                                                                             // 225
SignupKeys = new Mongo.Collection("signupKeys");                                                             // 226
// Invite keys which may be used by users to get access to Sandstorm.                                        // 227
//                                                                                                           // 228
// Each contains:                                                                                            // 229
//   _id:  random                                                                                            // 230
//   used:  Boolean indicating whether this key has already been consumed.                                   // 231
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.                // 232
//   email: If this key was sent as an email invite, the email address to which it was sent.                 // 233
                                                                                                             // 234
ActivityStats = new Mongo.Collection("activityStats");                                                       // 235
// Contains usage statistics taken on a regular interval. Each entry is a data point.                        // 236
//                                                                                                           // 237
// Each contains:                                                                                            // 238
//   timestamp: Date when measurements were taken.                                                           // 239
//   daily: Contains stats counts pertaining to the last day before the sample time.                         // 240
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.                 // 241
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.                 // 242
//                                                                                                           // 243
// Each of daily, weekly, and monthly contains:                                                              // 244
//   activeUsers: The number of unique users who have used a grain on the server in the time                 // 245
//       interval. Only counts logged-in users.                                                              // 246
//   demoUsers: Demo users.                                                                                  // 247
//   appDemoUsers: Users that came in through "app demo".                                                    // 248
//   activeGrains: The number of unique grains that have been used in the time interval.                     // 249
//   apps: An object indexed by app ID recording, for each app:                                              // 250
//       owners: Number of unique owners of this app (counting only grains that still exist).                // 251
//       sharedUsers: Number of users who have accessed other people's grains of this app (counting          // 252
//         only grains that still exist).                                                                    // 253
//       grains: Number of active grains of this app (that still exist).                                     // 254
//       deleted: Number of non-demo grains of this app that were deleted.                                   // 255
//       demoed: Number of demo grains created and expired.                                                  // 256
//       appDemoUsers: Number of app demos initiated with this app.                                          // 257
                                                                                                             // 258
DeleteStats = new Mongo.Collection("deleteStats");                                                           // 259
// Contains records of objects that were deleted, for stat-keeping purposes.                                 // 260
//                                                                                                           // 261
// Each contains:                                                                                            // 262
//   type: "grain" or "user" or "demoGrain" or "demoUser" or "appDemoUser"                                   // 263
//   lastActive: Date of the user's or grain's last activity.                                                // 264
//   appId: For type = "grain", the app ID of the grain. For type = "appDemoUser", the app ID they           // 265
//     arrived to demo. For others, undefined.                                                               // 266
                                                                                                             // 267
FileTokens = new Mongo.Collection("fileTokens");                                                             // 268
// Tokens corresponding to backup files that are currently stored on the server. A user receives             // 269
// a token when they create a backup file (either by uploading it, or by backing up one of their             // 270
// grains) and may use the token to read the file (either to download it, or to restore a new                // 271
// grain from it).                                                                                           // 272
//                                                                                                           // 273
// Each contains:                                                                                            // 274
//   _id:       The unguessable token string.                                                                // 275
//   name:      Suggested filename.                                                                          // 276
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.              // 277
                                                                                                             // 278
ApiTokens = new Mongo.Collection("apiTokens");                                                               // 279
// Access tokens for APIs exported by apps.                                                                  // 280
//                                                                                                           // 281
// Originally API tokens were only used by external users through the HTTP API endpoint. However,            // 282
// now they are also used to implement SturdyRefs, not just held by external users, but also when            // 283
// an app holds a SturdyRef to another app within the same server. See the various `save()`,                 // 284
// `restore()`, and `drop()` methods in `grain.capnp` (on `SansdtormApi`, `AppPersistent`, and               // 285
// `MainView`) -- the fields of type `Data` are API tokens.                                                  // 286
//                                                                                                           // 287
// Each contains:                                                                                            // 288
//   _id:       A SHA-256 hash of the token, base64-encoded.                                                 // 289
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)            // 290
//   identityId: For UiView capabilities, this is the identity for which the view is attenuated.             // 291
//              That is, the UiView's newSession() method will intersect the requested permissions           // 292
//              with this identity's permissions before forwarding on to the underlying app. If              // 293
//              `identityId` is not present, then no identity attenuation is applied, i.e. this is           // 294
//              a raw UiView as implemented by the app. (The `roleAssignment` field, below, may              // 295
//              still apply. For non-UiView capabilities, `identityId` is never present. Note that           // 296
//              this is NOT the identity against which the `requiredPermissions` parameter of                // 297
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerIdentity`.         // 298
//   accountId: For tokens where `identityId` is set, the `_id` (in the Users table) of the account          // 299
//              that created the token.                                                                      // 300
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded               // 301
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These          // 302
//              permissions will be intersected with those held by `identityId` when the view is             // 303
//              opened.                                                                                      // 304
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as          // 305
//              anonymous rather than as directly associated with `identityId`. This has no effect           // 306
//              on the permissions granted.                                                                  // 307
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by            // 308
//              the app or its supervisor (whereas without this it strictly represents UiView).              // 309
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).            // 310
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is                   // 311
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.                     // 312
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,          // 313
//              not a particular grain. (`grainId` and `identityId` are not set.) This is an object          // 314
//              containing exactly one of the following fields:                                              // 315
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                          // 316
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the                 // 317
//                           `Notifications` collection.                                                     // 318
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this            // 319
//                  will be moved out of the frontend and into the backend, but we'll migrate the            // 320
//                  database when that happens. This field contains the boolean true to signify that         // 321
//                  it has been set.                                                                         // 322
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                           // 323
//   parentToken: If present, then this token represents exactly the capability represented by               // 324
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated          // 325
//              by `roleAssignment` (if present). To facilitate permissions computations, if the             // 326
//              capability is a UiView, then `grainId` is set to the backing grain, `identityId`             // 327
//              is set to the identity that shared the view, and `accountId` is set to the account           // 328
//              that shared the view. Neither `objectId` nor `frontendRef` is present when                   // 329
//              `parentToken` is present.                                                                    // 330
//   petname:   Human-readable label for this access token, useful for identifying tokens for                // 331
//              revocation. This should be displayed when visualizing incoming capabilities to               // 332
//              the grain identified by `grainId`.                                                           // 333
//   created:   Date when this token was created.                                                            // 334
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may          // 335
//              become un-revoked in the future.                                                             // 336
//   expires:   Optional expiration Date. If undefined, the token does not expire.                           // 337
//   owner:     A `ApiTokenRefOwner` (defined in `supervisor.capnp`, stored as a JSON object)                // 338
//              as passed to the `save()` call that created this token. If not present, treat                // 339
//              as `webkey` (the default for `ApiTokenOwner`).                                               // 340
//   expiresIfUnused:                                                                                        // 341
//              Optional Date after which the token, if it has not been used yet, expires.                   // 342
//              This field should be cleared on a token's first use.                                         // 343
//   requirements: List of conditions which must hold for this token to be considered valid.                 // 344
//              Semantically, this list specifies the powers which were *used* to originally                 // 345
//              create the token. If any condition in the list becomes untrue, then the token must           // 346
//              be considered revoked, and all live refs and sturdy refs obtained transitively               // 347
//              through it must also become revoked. Each item is the JSON serialization of the              // 348
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                               // 349
//                                                                                                           // 350
// It is important to note that a token's owner and provider are independent from each other. To             // 351
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:          // 352
//                                                                                                           // 353
// struct ApiToken {                                                                                         // 354
//   owner :ApiTokenOwner;                                                                                   // 355
//   provider :union {                                                                                       // 356
//     grain :group {                                                                                        // 357
//       grainId :Text;                                                                                      // 358
//       union {                                                                                             // 359
//         uiView :group {                                                                                   // 360
//           identityId :Text;                                                                               // 361
//           roleAssignment :RoleAssignment;                                                                 // 362
//           forSharing :Bool;                                                                               // 363
//         }                                                                                                 // 364
//         objectId :SupervisorObjectId;                                                                     // 365
//       }                                                                                                   // 366
//     }                                                                                                     // 367
//     frontendRef :union {                                                                                  // 368
//        notificationHandle :Text;                                                                          // 369
//        ipNetwork :Bool;                                                                                   // 370
//        ipInterface :Bool;                                                                                 // 371
//     }                                                                                                     // 372
//     child :group {                                                                                        // 373
//       parentToken :Text;                                                                                  // 374
//       union {                                                                                             // 375
//         uiView :group {                                                                                   // 376
//           grainId :Text;                                                                                  // 377
//           identityId :Text;                                                                               // 378
//           roleAssignment :RoleAssignment = (allAccess = ());                                              // 379
//         }                                                                                                 // 380
//         other :Void;                                                                                      // 381
//       }                                                                                                   // 382
//     }                                                                                                     // 383
//   }                                                                                                       // 384
//   requirements: List(Supervisor.MembraneRequirement);                                                     // 385
//   ...                                                                                                     // 386
// }                                                                                                         // 387
                                                                                                             // 388
Notifications = new Mongo.Collection("notifications");                                                       // 389
// Notifications for a user.                                                                                 // 390
//                                                                                                           // 391
// Each contains:                                                                                            // 392
//   _id:          random                                                                                    // 393
//   grainId:      The grain originating this notification, if any.                                          // 394
//   userId:       The user receiving the notification.                                                      // 395
//   text:         The JSON-ified LocalizedText to display in the notification.                              // 396
//   isUnread:     Boolean indicating if this notification is unread.                                        // 397
//   timestamp:    Date when this notification was last updated                                              // 398
//   ongoing:      If present, this is an ongoing notification, and this field contains an                   // 399
//                 ApiToken referencing the `OngoingNotification` capability.                                // 400
//   admin:        If present, this is a notification intended for an admin.                                 // 401
//     action:     If present, this is a (string) link that the notification should direct the               // 402
//                 admin to.                                                                                 // 403
//     type:       The type of notification (currently only "reportStats").                                  // 404
//   appUpdates:   If present, this is an app update notification. It is an object with the appIds           // 405
//                 as keys.                                                                                  // 406
//     $appId:     The appId that has an outstanding update.                                                 // 407
//       packageId: The packageId that it will update to.                                                    // 408
//       name: The name of the app. (appTitle from package.manifest)                                         // 409
//       version: The app's version number. (appVersion from package.manifest)                               // 410
//       marketingVersion: String marketing version of this app. (appMarketingVersion from package.manifest)
                                                                                                             // 412
StatsTokens = new Mongo.Collection("statsTokens");                                                           // 413
// Access tokens for the Stats collection                                                                    // 414
//                                                                                                           // 415
// These tokens are used for accessing the ActivityStats collection remotely                                 // 416
// (ie. from a dashboard webapp)                                                                             // 417
//                                                                                                           // 418
// Each contains:                                                                                            // 419
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                        // 420
                                                                                                             // 421
Misc = new Mongo.Collection("misc");                                                                         // 422
// Miscellaneous configuration and other settings                                                            // 423
//                                                                                                           // 424
// This table is currently only used for persisting BASE_URL from one session to the next,                   // 425
// but in general any miscellaneous settings should go in here                                               // 426
//                                                                                                           // 427
// Each contains:                                                                                            // 428
//   _id:       The name of the setting. eg. "BASE_URL"                                                      // 429
//   value:     The value of the setting.                                                                    // 430
                                                                                                             // 431
Settings = new Mongo.Collection("settings");                                                                 // 432
// Settings for this Sandstorm instance go here. They are configured through the adminSettings               // 433
// route. This collection differs from misc in that this collection is completely user controlled.           // 434
//                                                                                                           // 435
// Each contains:                                                                                            // 436
//   _id:       The name of the setting. eg. "MAIL_URL"                                                      // 437
//   value:     The value of the setting.                                                                    // 438
//   potentially other fields that are unique to the setting                                                 // 439
                                                                                                             // 440
Migrations = new Mongo.Collection("migrations");                                                             // 441
// This table tracks which migrations we have applied to this instance.                                      // 442
// It contains a single entry:                                                                               // 443
//   _id:       "migrations_applied"                                                                         // 444
//   value:     The number of migrations this instance has successfully completed.                           // 445
                                                                                                             // 446
StaticAssets = new Mongo.Collection("staticAssets");                                                         // 447
// Collection of static assets served up from the Sandstorm server's "static" host. We only                  // 448
// support relatively small assets: under 1MB each.                                                          // 449
//                                                                                                           // 450
// Each contains:                                                                                            // 451
//   _id:       Random ID; will be used in the URL.                                                          // 452
//   hash:      A base64-encoded SHA-256 hash of the data, used to de-dupe.                                  // 453
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                                    // 454
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                          // 455
//   content:   The asset content (byte buffer).                                                             // 456
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't          // 457
//       have transactions, this needs to bias towards over-counting; a backup GC could be used              // 458
//       to catch leaked assets, although it's probably not a big deal in practice.                          // 459
                                                                                                             // 460
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                               // 461
// Collection of tokens representing a single-use permission to upload an asset, such as a new               // 462
// profile picture.                                                                                          // 463
//                                                                                                           // 464
// Each contains:                                                                                            // 465
//   _id:       Random ID.                                                                                   // 466
//   purpose:   Contains one of the following, indicating how the asset is to be used:                       // 467
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:                // 468
//           userId: User whose picture shall be replaced.                                                   // 469
//           identityId: Which of the user's identities shall be updated.                                    // 470
//   expires:   Time when this token will go away if unused.                                                 // 471
                                                                                                             // 472
Plans = new Mongo.Collection("plans");                                                                       // 473
// Subscription plans, which determine quota.                                                                // 474
//                                                                                                           // 475
// Each contains:                                                                                            // 476
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...                      // 477
//   storage: Number of bytes this user is allowed to store.                                                 // 478
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                                // 479
//   computeLabel: Label to display to the user describing this plan's compute units.                        // 480
//   grains: Total number of grains this user can create (often `Infinity`).                                 // 481
//   price: Price per month in US cents.                                                                     // 482
                                                                                                             // 483
AppIndex = new Mongo.Collection("appIndex");                                                                 // 484
// A mirror of the data from the App Market index                                                            // 485
//                                                                                                           // 486
// Each contains:                                                                                            // 487
//   _id: the appId of the app                                                                               // 488
//  The rest of the fields are defined in src/sandstorm/app-index/app-index.capnp:AppIndexForMarket          // 489
                                                                                                             // 490
KeybaseProfiles = new Mongo.Collection("keybaseProfiles");                                                   // 491
// Cache of Keybase profile information. The profile for a user is re-fetched every time a package           // 492
// by that user is installed, as well as if the keybase profile is requested and not already                 // 493
// present for some reason.                                                                                  // 494
//                                                                                                           // 495
// Each contains:                                                                                            // 496
//   _id: PGP key fingerprint (SHA-1, hex, all-caps)                                                         // 497
//   displayName: Display name from Keybase. (NOT VERIFIED AT ALL.)                                          // 498
//   handle: Keybase handle.                                                                                 // 499
//   proofs: The "proofs_summary.all" array from the Keybase lookup. See the non-existent Keybase            // 500
//     docs for details. We also add a boolean "status" field to each proof indicating whether               // 501
//     we have directly verified the proof ourselves. Its values may be "unverified" (Keybase                // 502
//     returned this but we haven't checked it directly), "verified" (we verified the proof and it           // 503
//     is valid), "invalid" (we checked the proof and it was definitely bogus), or "checking" (the           // 504
//     server is currently actively checking this proof). Note that if a check fails due to network          // 505
//     errors, the status goes back to "unverified".                                                         // 506
//                                                                                                           // 507
//     WARNING: Currently verification is NOT IMPLEMENTED, so all proofs will be "unverified"                // 508
//       for now and we just trust Keybase.                                                                  // 509
                                                                                                             // 510
if (Meteor.isServer) {                                                                                       // 511
  Meteor.publish("credentials", function () {                                                                // 512
    // Data needed for isSignedUp() and isAdmin() to work.                                                   // 513
                                                                                                             // 514
    if (this.userId) {                                                                                       // 515
      return [                                                                                               // 516
        Meteor.users.find({_id: this.userId},                                                                // 517
            {fields: {signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,                                 // 518
                      plan: 1, hasCompletedSignup: 1}}),                                                     // 519
        Plans.find()                                                                                         // 520
      ];                                                                                                     // 521
    } else {                                                                                                 // 522
      return [];                                                                                             // 523
    }                                                                                                        // 524
  });                                                                                                        // 525
}                                                                                                            // 526
                                                                                                             // 527
isDemoUser = function() {                                                                                    // 528
  // Returns true if this is a demo user.                                                                    // 529
                                                                                                             // 530
  var user = Meteor.user();                                                                                  // 531
  if (user && user.expires) {                                                                                // 532
    return true;                                                                                             // 533
  } else {                                                                                                   // 534
    return false;                                                                                            // 535
  }                                                                                                          // 536
}                                                                                                            // 537
                                                                                                             // 538
isSignedUp = function() {                                                                                    // 539
  // Returns true if the user has presented an invite key.                                                   // 540
                                                                                                             // 541
  var user = Meteor.user();                                                                                  // 542
                                                                                                             // 543
  if (!user) return false;  // not signed in                                                                 // 544
                                                                                                             // 545
  if (user.expires) return false;  // demo user.                                                             // 546
                                                                                                             // 547
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 548
                                                                                                             // 549
  if (user.signupKey) return true;  // user is invited                                                       // 550
                                                                                                             // 551
  return false;                                                                                              // 552
}                                                                                                            // 553
                                                                                                             // 554
isSignedUpOrDemo = function () {                                                                             // 555
  var user = Meteor.user();                                                                                  // 556
                                                                                                             // 557
  if (!user) return false;  // not signed in                                                                 // 558
                                                                                                             // 559
  if (user.expires) return true;  // demo user.                                                              // 560
                                                                                                             // 561
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                           // 562
                                                                                                             // 563
  if (user.signupKey) return true;  // user is invited                                                       // 564
                                                                                                             // 565
  return false;                                                                                              // 566
}                                                                                                            // 567
                                                                                                             // 568
var calculateReferralBonus = function(accountId, plan) {                                                     // 569
  // This function returns an object of the form:                                                            // 570
  //                                                                                                         // 571
  // - {grains: 0, storage: 0}                                                                               // 572
  //                                                                                                         // 573
  // which are extra resources this account gets as part of participating in the referral                    // 574
  // program. (Storage is measured in bytes, as usual for plans.)                                            // 575
                                                                                                             // 576
                                                                                                             // 577
  // Authorization note: Only call this if accountId is the current user!                                    // 578
  var isPaid = (plan && plan !== "free");                                                                    // 579
                                                                                                             // 580
  successfulReferralsCount = countReferrals(accountId);                                                      // 581
  if (isPaid) {                                                                                              // 582
    var maxPaidStorageBonus = 30 * 1e9;                                                                      // 583
    return {grains: 0,                                                                                       // 584
            storage: Math.min(                                                                               // 585
              successfulReferralsCount * 2 * 1e9,                                                            // 586
              maxPaidStorageBonus)};                                                                         // 587
  } else {                                                                                                   // 588
    var maxFreeStorageBonus = 2 * 1e9;                                                                       // 589
    var bonus = {                                                                                            // 590
      storage: Math.min(                                                                                     // 591
        successfulReferralsCount * 50 * 1e6,                                                                 // 592
        maxFreeStorageBonus)                                                                                 // 593
    };                                                                                                       // 594
    if (successfulReferralsCount > 0) {                                                                      // 595
      bonus.grains = Infinity;                                                                               // 596
    } else {                                                                                                 // 597
      bonus.grains = 0;                                                                                      // 598
    }                                                                                                        // 599
    return bonus;                                                                                            // 600
  }                                                                                                          // 601
}                                                                                                            // 602
                                                                                                             // 603
var countReferrals = function (accountId) {                                                                  // 604
  var referredIdentityIds = Meteor.users.findOne(                                                            // 605
    {_id: accountId},                                                                                        // 606
    {fields: {referredIdentityIds: 1}}).referredIdentityIds;                                                 // 607
  return (referredIdentityIds && referredIdentityIds.length || 0);                                           // 608
}                                                                                                            // 609
                                                                                                             // 610
getUserQuota = function (user) {                                                                             // 611
  var plan = Plans.findOne(user.plan || "free");                                                             // 612
  var referralBonus = calculateReferralBonus(user._id, plan);                                                // 613
  var userQuota = {};                                                                                        // 614
  userQuota.storage = (plan.storage + referralBonus.storage);                                                // 615
  userQuota.grains = (plan.grains + referralBonus.grains);                                                   // 616
  return userQuota;                                                                                          // 617
}                                                                                                            // 618
                                                                                                             // 619
isUserOverQuota = function (user) {                                                                          // 620
  // Return false if user has quota space remaining, true if it is full. When this returns true,             // 621
  // we will not allow the user to create new grains, though they may be able to open existing ones          // 622
  // which may still increase their storage usage.                                                           // 623
  //                                                                                                         // 624
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 625
                                                                                                             // 626
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 627
                                                                                                             // 628
  var plan = getUserQuota(user);                                                                             // 629
  if (plan.grains < Infinity) {                                                                              // 630
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains}).count();                   // 631
    if (count >= plan.grains) return "outOfGrains";                                                          // 632
  }                                                                                                          // 633
                                                                                                             // 634
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";                   // 635
}                                                                                                            // 636
                                                                                                             // 637
isUserExcessivelyOverQuota = function (user) {                                                               // 638
  // Return true if user is so far over quota that we should prevent their existing grains from              // 639
  // running at all.                                                                                         // 640
  //                                                                                                         // 641
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                        // 642
                                                                                                             // 643
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                                    // 644
                                                                                                             // 645
  var quota = getUserQuota(user);                                                                            // 646
                                                                                                             // 647
  // quota.grains = Infinity means unlimited grains. IEEE754 defines Infinity == Infinity.                   // 648
  if (quota.grains < Infinity) {                                                                             // 649
    var count = Grains.find({userId: user._id}, {fields: {}, limit: quota.grains * 2}).count();              // 650
    if (count >= quota.grains * 2) return "outOfGrains";                                                     // 651
  }                                                                                                          // 652
                                                                                                             // 653
  return quota && user.storageUsage && user.storageUsage >= quota.storage * 1.2 && "outOfStorage";           // 654
}                                                                                                            // 655
                                                                                                             // 656
isAdmin = function() {                                                                                       // 657
  // Returns true if the user is the administrator.                                                          // 658
                                                                                                             // 659
  var user = Meteor.user();                                                                                  // 660
  if (user && user.isAdmin) {                                                                                // 661
    return true;                                                                                             // 662
  } else {                                                                                                   // 663
    return false;                                                                                            // 664
  }                                                                                                          // 665
}                                                                                                            // 666
                                                                                                             // 667
isAdminById = function(id) {                                                                                 // 668
  // Returns true if the user's id is the administrator.                                                     // 669
                                                                                                             // 670
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                         // 671
  if (user && user.isAdmin) {                                                                                // 672
    return true;                                                                                             // 673
  } else {                                                                                                   // 674
    return false;                                                                                            // 675
  }                                                                                                          // 676
}                                                                                                            // 677
                                                                                                             // 678
findAdminUserForToken = function (token) {                                                                   // 679
  if (!token.requirements) {                                                                                 // 680
    return;                                                                                                  // 681
  }                                                                                                          // 682
  var requirements = token.requirements.filter(function (requirement) {                                      // 683
    return "userIsAdmin" in requirement;                                                                     // 684
  });                                                                                                        // 685
                                                                                                             // 686
  if (requirements.length > 1) {                                                                             // 687
    return;                                                                                                  // 688
  }                                                                                                          // 689
  if (requirements.length === 0) {                                                                           // 690
    return;                                                                                                  // 691
  }                                                                                                          // 692
  return requirements[0].userIsAdmin;                                                                        // 693
};                                                                                                           // 694
                                                                                                             // 695
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                             // 696
                                                                                                             // 697
if (wildcardHost.length != 2) {                                                                              // 698
  throw new Error("Wildcard host must contain exactly one asterisk.");                                       // 699
}                                                                                                            // 700
                                                                                                             // 701
matchWildcardHost = function(host) {                                                                         // 702
  // See if the hostname is a member of our wildcard. If so, extract the ID.                                 // 703
                                                                                                             // 704
  var prefix = wildcardHost[0];                                                                              // 705
  var suffix = wildcardHost[1];                                                                              // 706
                                                                                                             // 707
  // We remove everything after the first ":" character so that our                                          // 708
  // comparison logic ignores port numbers.                                                                  // 709
  suffix = suffix.split(":")[0];                                                                             // 710
  host = host.split(":")[0];                                                                                 // 711
                                                                                                             // 712
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                                    // 713
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                           // 714
      host.length >= prefix.length + suffix.length) {                                                        // 715
    var id = host.slice(prefix.length, -suffix.length);                                                      // 716
    if (id.match(/^[a-z0-9]*$/)) {                                                                           // 717
      return id;                                                                                             // 718
    }                                                                                                        // 719
  }                                                                                                          // 720
                                                                                                             // 721
  return null;                                                                                               // 722
}                                                                                                            // 723
                                                                                                             // 724
makeWildcardHost = function (id) {                                                                           // 725
  return wildcardHost[0] + id + wildcardHost[1];                                                             // 726
}                                                                                                            // 727
                                                                                                             // 728
if (Meteor.isServer) {                                                                                       // 729
  var Url = Npm.require("url");                                                                              // 730
  getWildcardOrigin = function () {                                                                          // 731
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're              // 732
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),            // 733
    // an astrisk is only allowed as the first character and must be followed by a period. So we need        // 734
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can           // 735
    // really do. We also add the protocol to the front (again, that's what CSP wants).                      // 736
                                                                                                             // 737
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.         // 738
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                                 // 739
                                                                                                             // 740
    var dotPos = wildcardHost[1].indexOf(".");                                                               // 741
    if (dotPos < 0) {                                                                                        // 742
      return protocol + "//*";                                                                               // 743
    } else {                                                                                                 // 744
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                               // 745
    }                                                                                                        // 746
  }                                                                                                          // 747
}                                                                                                            // 748
                                                                                                             // 749
allowDevAccounts = function () {                                                                             // 750
  var setting = Settings.findOne({_id: "devAccounts"});                                                      // 751
  if (setting) {                                                                                             // 752
    return setting.value;                                                                                    // 753
  } else {                                                                                                   // 754
    return Meteor.settings && Meteor.settings.public &&                                                      // 755
           Meteor.settings.public.allowDevAccounts;                                                          // 756
  }                                                                                                          // 757
};                                                                                                           // 758
                                                                                                             // 759
roleAssignmentPattern = {                                                                                    // 760
  none : Match.Optional(null),                                                                               // 761
  allAccess: Match.Optional(null),                                                                           // 762
  roleId: Match.Optional(Match.Integer),                                                                     // 763
  addPermissions: Match.Optional([Boolean]),                                                                 // 764
  removePermissions: Match.Optional([Boolean]),                                                              // 765
};                                                                                                           // 766
                                                                                                             // 767
SandstormDb = function () {                                                                                  // 768
  this.collections = {                                                                                       // 769
    // Direct access to underlying collections. DEPRECATED.                                                  // 770
    //                                                                                                       // 771
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove            // 772
    //   direct access to the collections.                                                                   // 773
                                                                                                             // 774
    packages: Packages,                                                                                      // 775
    devPackages: DevPackages,                                                                                // 776
    userActions: UserActions,                                                                                // 777
    grains: Grains,                                                                                          // 778
    contacts: Contacts,                                                                                      // 779
    sessions: Sessions,                                                                                      // 780
    signupKeys: SignupKeys,                                                                                  // 781
    activityStats: ActivityStats,                                                                            // 782
    deleteStats: DeleteStats,                                                                                // 783
    fileTokens: FileTokens,                                                                                  // 784
    apiTokens: ApiTokens,                                                                                    // 785
    notifications: Notifications,                                                                            // 786
    statsTokens: StatsTokens,                                                                                // 787
    misc: Misc,                                                                                              // 788
    settings: Settings,                                                                                      // 789
    appIndex: AppIndex,                                                                                      // 790
    keybaseProfiles: KeybaseProfiles,                                                                        // 791
                                                                                                             // 792
    // Intentionally omitted:                                                                                // 793
    // - Migrations, since it's used only within this package.                                               // 794
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.          // 795
  };                                                                                                         // 796
};                                                                                                           // 797
                                                                                                             // 798
// TODO(cleanup): These methods should not be defined freestanding and should use collection                 // 799
//   objects created in SandstormDb's constructor rather than globals.                                       // 800
                                                                                                             // 801
_.extend(SandstormDb.prototype, {                                                                            // 802
  isDemoUser: isDemoUser,                                                                                    // 803
  isSignedUp: isSignedUp,                                                                                    // 804
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                        // 805
  isUserOverQuota: isUserOverQuota,                                                                          // 806
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                                    // 807
  isAdmin: isAdmin,                                                                                          // 808
  isAdminById: isAdminById,                                                                                  // 809
  findAdminUserForToken: findAdminUserForToken,                                                              // 810
  matchWildcardHost: matchWildcardHost,                                                                      // 811
  makeWildcardHost: makeWildcardHost,                                                                        // 812
  allowDevAccounts: allowDevAccounts,                                                                        // 813
  roleAssignmentPattern: roleAssignmentPattern,                                                              // 814
});                                                                                                          // 815
                                                                                                             // 816
if (Meteor.isServer) {                                                                                       // 817
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                               // 818
}                                                                                                            // 819
                                                                                                             // 820
// =======================================================================================                   // 821
// Below this point are newly-written or refactored functions.                                               // 822
                                                                                                             // 823
_.extend(SandstormDb.prototype, {                                                                            // 824
  getUser: function getUser (userId) {                                                                       // 825
    check(userId, Match.OneOf(String, undefined, null));                                                     // 826
    if (userId) {                                                                                            // 827
      return Meteor.users.findOne(userId);                                                                   // 828
    }                                                                                                        // 829
  },                                                                                                         // 830
                                                                                                             // 831
  getIdentity: function getIdentity (identityId) {                                                           // 832
    check(identityId, String);                                                                               // 833
    var identity = Meteor.users.findOne({_id: identityId});                                                  // 834
    if (identity) {                                                                                          // 835
      SandstormDb.fillInProfileDefaults(identity);                                                           // 836
      SandstormDb.fillInIntrinsicName(identity);                                                             // 837
      SandstormDb.fillInPictureUrl(identity);                                                                // 838
      return identity;                                                                                       // 839
    }                                                                                                        // 840
  },                                                                                                         // 841
                                                                                                             // 842
  userHasIdentity: function (userId, identityId) {                                                           // 843
    check(userId, String);                                                                                   // 844
    check(identityId, String);                                                                               // 845
                                                                                                             // 846
    if (userId === identityId) return true;                                                                  // 847
                                                                                                             // 848
    var user = Meteor.users.findOne(userId);                                                                 // 849
    return SandstormDb.getUserIdentityIds(user).indexOf(identityId) != -1;                                   // 850
  },                                                                                                         // 851
                                                                                                             // 852
  userGrains: function userGrains (userId) {                                                                 // 853
    check(userId, Match.OneOf(String, undefined, null));                                                     // 854
    return this.collections.grains.find({userId: userId});                                                   // 855
  },                                                                                                         // 856
                                                                                                             // 857
  currentUserGrains: function currentUserGrains () {                                                         // 858
    return this.userGrains(Meteor.userId());                                                                 // 859
  },                                                                                                         // 860
                                                                                                             // 861
  getGrain: function getGrain (grainId) {                                                                    // 862
    check(grainId, String);                                                                                  // 863
    return this.collections.grains.findOne(grainId);                                                         // 864
  },                                                                                                         // 865
                                                                                                             // 866
  userApiTokens: function userApiTokens (userId) {                                                           // 867
    check(userId, Match.OneOf(String, undefined, null));                                                     // 868
    var identityIds = SandstormDb.getUserIdentityIds(this.getUser(userId));                                  // 869
    return this.collections.apiTokens.find({'owner.user.identityId': {$in: identityIds}});                   // 870
  },                                                                                                         // 871
                                                                                                             // 872
  currentUserApiTokens: function currentUserApiTokens () {                                                   // 873
    return this.userApiTokens(Meteor.userId());                                                              // 874
  },                                                                                                         // 875
                                                                                                             // 876
  userActions: function userActions (user) {                                                                 // 877
    return this.collections.userActions.find({userId: user});                                                // 878
  },                                                                                                         // 879
                                                                                                             // 880
  currentUserActions: function currentUserActions () {                                                       // 881
    return this.userActions(Meteor.userId());                                                                // 882
  },                                                                                                         // 883
                                                                                                             // 884
  iconSrcForPackage: function iconSrcForPackage (pkg, usage) {                                               // 885
    return Identicon.iconSrcForPackage(pkg, usage, this.makeWildcardHost("static"));                         // 886
  },                                                                                                         // 887
                                                                                                             // 888
  getPlan: function (id) {                                                                                   // 889
    check(id, String);                                                                                       // 890
    var plan = Plans.findOne(id);                                                                            // 891
    if (!plan) {                                                                                             // 892
      throw new Error("no such plan: " + id);                                                                // 893
    }                                                                                                        // 894
    return plan;                                                                                             // 895
  },                                                                                                         // 896
                                                                                                             // 897
  listPlans: function () {                                                                                   // 898
    return Plans.find({}, {sort: {price: 1}});                                                               // 899
  },                                                                                                         // 900
                                                                                                             // 901
  getMyPlan: function () {                                                                                   // 902
    var user = Meteor.user();                                                                                // 903
    return user && Plans.findOne(user.plan || "free");                                                       // 904
  },                                                                                                         // 905
                                                                                                             // 906
  getMyReferralBonus: function(user) {                                                                       // 907
    // This function is called from the server and from the client, similar to getMyPlan().                  // 908
    //                                                                                                       // 909
    // When called from the server, calculate the user's actual referral bonus. We use this                  // 910
    // elsewhere to store a value in user.pseudoReferralBonus.                                               // 911
    //                                                                                                       // 912
    // When called from the client, return the value of user.pseudoReferralBonus (if it exists).             // 913
    if (Meteor.isClient) {                                                                                   // 914
      // If called on the client side, always use the currently logged-in user.                              // 915
      user = Meteor.user();                                                                                  // 916
                                                                                                             // 917
      if (user && user.pseudoReferralBonus) {                                                                // 918
        return user.pseudoReferralBonus;                                                                     // 919
      }                                                                                                      // 920
      // If we get to this, the subscriptions haven't arrived yet.                                           // 921
      var noBonus = {grains: 0, storage: 0};                                                                 // 922
      return noBonus;                                                                                        // 923
    }                                                                                                        // 924
    if (Meteor.isServer) {                                                                                   // 925
      var x = calculateReferralBonus(user._id, user.plan);                                                   // 926
      return x;                                                                                              // 927
    }                                                                                                        // 928
  },                                                                                                         // 929
                                                                                                             // 930
  getMyUsage: function (user) {                                                                              // 931
    user = user || Meteor.user();                                                                            // 932
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                                     // 933
      if (Meteor.isClient) {                                                                                 // 934
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly              // 935
        // not reactive.                                                                                     // 936
        return user.pseudoUsage;                                                                             // 937
      } else {                                                                                               // 938
        return {                                                                                             // 939
          grains: Grains.find({userId: user._id}).count(),                                                   // 940
          storage: user.storageUsage || 0,                                                                   // 941
          compute: 0   // not tracked yet                                                                    // 942
        };                                                                                                   // 943
      }                                                                                                      // 944
    } else {                                                                                                 // 945
      return {grains: 0, storage: 0, compute: 0};                                                            // 946
    }                                                                                                        // 947
  },                                                                                                         // 948
                                                                                                             // 949
  isUninvitedFreeUser: function () {                                                                         // 950
    if (!Meteor.settings.public.allowUninvited) return false;                                                // 951
                                                                                                             // 952
    var user = Meteor.user();                                                                                // 953
    return user && !user.expires && (!user.plan || user.plan === "free");                                    // 954
  },                                                                                                         // 955
                                                                                                             // 956
  getSetting: function (name) {                                                                              // 957
    var setting = Settings.findOne(name);                                                                    // 958
    return setting && setting.value;                                                                         // 959
  },                                                                                                         // 960
                                                                                                             // 961
  addUserActions: function(packageId) {                                                                      // 962
    //TODO(cleanup): implement this with meteor methods rather than client-side inserts/removes.             // 963
    var pack = Packages.findOne(packageId);                                                                  // 964
    if (pack) {                                                                                              // 965
      // Remove old versions.                                                                                // 966
      UserActions.find({userId: Meteor.userId(), appId: pack.appId})                                         // 967
          .forEach(function (action) {                                                                       // 968
        UserActions.remove(action._id);                                                                      // 969
      });                                                                                                    // 970
                                                                                                             // 971
      // Install new.                                                                                        // 972
      var actions = pack.manifest.actions;                                                                   // 973
      for (i in actions) {                                                                                   // 974
        var action = actions[i];                                                                             // 975
        if ("none" in action.input) {                                                                        // 976
          var userAction = {                                                                                 // 977
            userId: Meteor.userId(),                                                                         // 978
            packageId: pack._id,                                                                             // 979
            appId: pack.appId,                                                                               // 980
            appTitle: pack.manifest.appTitle,                                                                // 981
            appMarketingVersion: pack.manifest.appMarketingVersion,                                          // 982
            appVersion: pack.manifest.appVersion,                                                            // 983
            title: action.title,                                                                             // 984
            nounPhrase: action.nounPhrase,                                                                   // 985
            command: action.command                                                                          // 986
          };                                                                                                 // 987
          UserActions.insert(userAction);                                                                    // 988
        } else {                                                                                             // 989
          // TODO(someday):  Implement actions with capability inputs.                                       // 990
        }                                                                                                    // 991
      }                                                                                                      // 992
                                                                                                             // 993
      Meteor.call("deleteUnusedPackages", pack.appId);                                                       // 994
    }                                                                                                        // 995
  },                                                                                                         // 996
                                                                                                             // 997
  sendAdminNotification: function (message, link) {                                                          // 998
    Meteor.users.find({isAdmin: true}, {fields: {_id: 1}}).forEach(function (user) {                         // 999
      Notifications.insert({                                                                                 // 1000
        admin: {                                                                                             // 1001
          action: link,                                                                                      // 1002
          type: "reportStats",                                                                               // 1003
        },                                                                                                   // 1004
        userId: user._id,                                                                                    // 1005
        text: {defaultText: message},                                                                        // 1006
        timestamp: new Date(),                                                                               // 1007
        isUnread: true,                                                                                      // 1008
      });                                                                                                    // 1009
    });                                                                                                      // 1010
  },                                                                                                         // 1011
                                                                                                             // 1012
  getKeybaseProfile: function (keyFingerprint) {                                                             // 1013
    return this.collections.keybaseProfiles.findOne(keyFingerprint) || {};                                   // 1014
  },                                                                                                         // 1015
});                                                                                                          // 1016
                                                                                                             // 1017
var appNameFromPackage = function(packageObj) {                                                              // 1018
  // This function takes a Package object from Mongo and returns an                                          // 1019
  // app title.                                                                                              // 1020
  var manifest = packageObj.manifest;                                                                        // 1021
  if (!manifest) return packageObj.appId || packageObj._id || "unknown";                                     // 1022
  var action = manifest.actions[0];                                                                          // 1023
  appName = (manifest.appTitle && manifest.appTitle.defaultText) ||                                          // 1024
    appNameFromActionName(action.title.defaultText);                                                         // 1025
  return appName;                                                                                            // 1026
};                                                                                                           // 1027
                                                                                                             // 1028
var appNameFromActionName = function(name) {                                                                 // 1029
  // Hack: Historically we only had action titles, like "New Etherpad Document", not app                     // 1030
  //   titles. But for this UI we want app titles. As a transitionary measure, try to                        // 1031
  //   derive the app title from the action title.                                                           // 1032
  // TODO(cleanup): Get rid of this once apps have real titles.                                              // 1033
  if (!name) {                                                                                               // 1034
    return "(unnamed)";                                                                                      // 1035
  }                                                                                                          // 1036
  if (name.lastIndexOf("New ", 0) === 0) {                                                                   // 1037
    name = name.slice(4);                                                                                    // 1038
  }                                                                                                          // 1039
  if (name.lastIndexOf("Hacker CMS", 0) === 0) {                                                             // 1040
    name = "Hacker CMS";                                                                                     // 1041
  } else {                                                                                                   // 1042
    var space = name.indexOf(" ");                                                                           // 1043
    if (space > 0) {                                                                                         // 1044
      name = name.slice(0, space);                                                                           // 1045
    }                                                                                                        // 1046
  }                                                                                                          // 1047
  return name;                                                                                               // 1048
};                                                                                                           // 1049
                                                                                                             // 1050
var appShortDescriptionFromPackage = function (pkg) {                                                        // 1051
  return pkg && pkg.manifest && pkg.manifest.metadata &&                                                     // 1052
         pkg.manifest.metadata.shortDescription &&                                                           // 1053
         pkg.manifest.metadata.shortDescription.defaultText;                                                 // 1054
};                                                                                                           // 1055
                                                                                                             // 1056
var nounPhraseForActionAndAppTitle = function(action, appTitle) {                                            // 1057
  // A hack to deal with legacy apps not including fields in their manifests.                                // 1058
  // I look forward to the day I can remove most of this code.                                               // 1059
  // Attempt to figure out the appropriate noun that this action will create.                                // 1060
  // Use an explicit noun phrase is one is available.  Apps should add these in the future.                  // 1061
  if (action.nounPhrase) return action.nounPhrase.defaultText;                                               // 1062
  // Otherwise, try to guess one from the structure of the action title field                                // 1063
  if (action.title && action.title.defaultText) {                                                            // 1064
    var text = action.title.defaultText;                                                                     // 1065
    // Strip a leading "New "                                                                                // 1066
    if (text.lastIndexOf("New ", 0) === 0) {                                                                 // 1067
      var candidate = text.slice(4);                                                                         // 1068
      // Strip a leading appname too, if provided                                                            // 1069
      if (candidate.lastIndexOf(appTitle, 0) === 0) {                                                        // 1070
        var newCandidate = candidate.slice(appTitle.length);                                                 // 1071
        // Unless that leaves you with no noun, in which case, use "grain"                                   // 1072
        if (newCandidate.length > 0) {                                                                       // 1073
          return newCandidate.toLowerCase();                                                                 // 1074
        } else {                                                                                             // 1075
          return "grain";                                                                                    // 1076
        }                                                                                                    // 1077
      }                                                                                                      // 1078
      return candidate.toLowerCase();                                                                        // 1079
    }                                                                                                        // 1080
    // Some other verb phrase was given.  Just use it verbatim, and hope the app author updates              // 1081
    // the package soon.                                                                                     // 1082
    return text;                                                                                             // 1083
  } else {                                                                                                   // 1084
    return "grain";                                                                                          // 1085
  }                                                                                                          // 1086
};                                                                                                           // 1087
                                                                                                             // 1088
// Static methods on SandstormDb that don't need an instance.                                                // 1089
// Largely things that deal with backwards-compatibility.                                                    // 1090
_.extend(SandstormDb, {                                                                                      // 1091
  appNameFromActionName: appNameFromActionName,                                                              // 1092
  appNameFromPackage: appNameFromPackage,                                                                    // 1093
  appShortDescriptionFromPackage: appShortDescriptionFromPackage,                                            // 1094
  nounPhraseForActionAndAppTitle: nounPhraseForActionAndAppTitle,                                            // 1095
});                                                                                                          // 1096
                                                                                                             // 1097
if (Meteor.isServer) {                                                                                       // 1098
  var Crypto = Npm.require("crypto");                                                                        // 1099
  var ContentType = Npm.require("content-type");                                                             // 1100
  var Zlib = Npm.require("zlib");                                                                            // 1101
                                                                                                             // 1102
  var replicaNumber = Meteor.settings.replicaNumber || 0;                                                    // 1103
                                                                                                             // 1104
  var computeStagger = function (n) {                                                                        // 1105
    // Compute a fraction in the range [0, 1) such that, for any natural number k, the values                // 1106
    // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.                 // 1107
    // The sequence looks like:                                                                              // 1108
    //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                                     // 1109
    //                                                                                                       // 1110
    // We use this to determine how we'll stagger periodic events performed by this replica.                 // 1111
    // Notice that this allows us to compute a stagger which is independent of the number of                 // 1112
    // front-end replicas present; we can add more replicas to the end without affecting how the             // 1113
    // earlier ones schedule their events.                                                                   // 1114
    var denom = 1;                                                                                           // 1115
    while (denom <= n) denom <<= 1;                                                                          // 1116
    var num = n * 2 - denom + 1;                                                                             // 1117
    return num / denom;                                                                                      // 1118
  }                                                                                                          // 1119
                                                                                                             // 1120
  var stagger = computeStagger(replicaNumber);                                                               // 1121
                                                                                                             // 1122
  SandstormDb.periodicCleanup = function (intervalMs, callback) {                                            // 1123
    // Register a database cleanup function than should run periodically, roughly once every                 // 1124
    // interval of the given length.                                                                         // 1125
    //                                                                                                       // 1126
    // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be              // 1127
    // scaled appropriately on the assumption that more data is being generated demanding more               // 1128
    // frequent cleanups.                                                                                    // 1129
                                                                                                             // 1130
    check(intervalMs, Number);                                                                               // 1131
    check(callback, Function);                                                                               // 1132
                                                                                                             // 1133
    if (intervalMs < 120000) {                                                                               // 1134
      throw new Error("less than 2-minute cleanup interval seems too fast; " +                               // 1135
                      "are you using the right units?");                                                     // 1136
    }                                                                                                        // 1137
                                                                                                             // 1138
    // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that              // 1139
    // the schedule is independent of the exact startup time.                                                // 1140
    var first = intervalMs - Date.now() % intervalMs;                                                        // 1141
                                                                                                             // 1142
    // Stagger cleanups across replicas so that we don't have all replicas trying to clean the               // 1143
    // same data at the same time.                                                                           // 1144
    first += Math.floor(intervalMs * computeStagger(replicaNumber));                                         // 1145
                                                                                                             // 1146
    // If the stagger put us more than an interval away from now, back up.                                   // 1147
    if (first > intervalMs) first -= intervalMs;                                                             // 1148
                                                                                                             // 1149
    Meteor.setTimeout(function () {                                                                          // 1150
      callback();                                                                                            // 1151
      Meteor.setInterval(callback, intervalMs);                                                              // 1152
    }, first);                                                                                               // 1153
  }                                                                                                          // 1154
                                                                                                             // 1155
  // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.                  // 1156
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                          // 1157
                                                                                                             // 1158
  var BufferSmallerThan = function (limit) {                                                                 // 1159
    return Match.Where(function (buf) {                                                                      // 1160
      check(buf, Buffer);                                                                                    // 1161
      return buf.length < limit;                                                                             // 1162
    });                                                                                                      // 1163
  }                                                                                                          // 1164
                                                                                                             // 1165
  var DatabaseId = Match.Where(function (s) {                                                                // 1166
    check(s, String);                                                                                        // 1167
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                                     // 1168
  });                                                                                                        // 1169
                                                                                                             // 1170
  addStaticAsset = function (metadata, content) {                                                            // 1171
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it             // 1172
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.          // 1173
                                                                                                             // 1174
    if (typeof content === "string" && !metadata.encoding) {                                                 // 1175
      content = gzipSync(new Buffer(content, "utf8"));                                                       // 1176
      metadata.encoding = "gzip";                                                                            // 1177
    }                                                                                                        // 1178
                                                                                                             // 1179
    check(metadata, {                                                                                        // 1180
      mimeType: String,                                                                                      // 1181
      encoding: Match.Optional("gzip")                                                                       // 1182
    });                                                                                                      // 1183
    check(content, BufferSmallerThan(1 << 20));                                                              // 1184
                                                                                                             // 1185
    // Validate content type.                                                                                // 1186
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                            // 1187
                                                                                                             // 1188
    var hasher = Crypto.createHash("sha256");                                                                // 1189
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                              // 1190
    hasher.update(content);                                                                                  // 1191
    var hash = hasher.digest("base64");                                                                      // 1192
                                                                                                             // 1193
    var existing = StaticAssets.findAndModify({                                                              // 1194
      query: {hash: hash, refcount: {$gte: 1}},                                                              // 1195
      update: {$inc: {refcount: 1}},                                                                         // 1196
      fields: {_id: 1, refcount: 1},                                                                         // 1197
    });                                                                                                      // 1198
    if (existing) {                                                                                          // 1199
      return existing._id;                                                                                   // 1200
    }                                                                                                        // 1201
                                                                                                             // 1202
    return StaticAssets.insert(_.extend({                                                                    // 1203
      hash: hash,                                                                                            // 1204
      content: content,                                                                                      // 1205
      refcount: 1                                                                                            // 1206
    }, metadata));                                                                                           // 1207
  }                                                                                                          // 1208
                                                                                                             // 1209
  SandstormDb.prototype.addStaticAsset = addStaticAsset;                                                     // 1210
                                                                                                             // 1211
  SandstormDb.prototype.refStaticAsset = function (id) {                                                     // 1212
    // Increment the refcount on an existing static asset.                                                   // 1213
    //                                                                                                       // 1214
    // You must call this BEFORE adding the new reference to the DB, in case of failure between              // 1215
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1216
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1217
                                                                                                             // 1218
    check(id, String);                                                                                       // 1219
                                                                                                             // 1220
    var existing = StaticAssets.findAndModify({                                                              // 1221
      query: {hash: hash},                                                                                   // 1222
      update: {$inc: {refcount: 1}},                                                                         // 1223
      fields: {_id: 1, refcount: 1},                                                                         // 1224
    });                                                                                                      // 1225
    if (!existing) {                                                                                         // 1226
      throw new Error("refStaticAsset() called on asset that doesn't exist");                                // 1227
    }                                                                                                        // 1228
  }                                                                                                          // 1229
                                                                                                             // 1230
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                                   // 1231
    // Decrement refcount on a static asset and delete if it has reached zero.                               // 1232
    //                                                                                                       // 1233
    // You must call this AFTER removing the reference from the DB, in case of failure between               // 1234
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big             // 1235
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                               // 1236
                                                                                                             // 1237
    check(id, String);                                                                                       // 1238
                                                                                                             // 1239
    var existing = StaticAssets.findAndModify({                                                              // 1240
      query: {_id: id},                                                                                      // 1241
      update: {$inc: {refcount: -1}},                                                                        // 1242
      fields: {_id: 1, refcount: 1},                                                                         // 1243
      new: true,                                                                                             // 1244
    });                                                                                                      // 1245
    if (!existing) {                                                                                         // 1246
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);               // 1247
    } else if (existing.refcount <= 0) {                                                                     // 1248
      StaticAssets.remove({_id: existing._id});                                                              // 1249
    }                                                                                                        // 1250
  }                                                                                                          // 1251
                                                                                                             // 1252
  SandstormDb.prototype.getStaticAsset = function (id) {                                                     // 1253
    // Get a static asset's mimeType, encoding, and raw content.                                             // 1254
                                                                                                             // 1255
    check(id, String);                                                                                       // 1256
                                                                                                             // 1257
    var asset = StaticAssets.findOne(id, {fields: {_id: 0, mimeType: 1, encoding: 1, content: 1}});          // 1258
    if (asset) {                                                                                             // 1259
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy              // 1260
      //   here.                                                                                             // 1261
      asset.content = new Buffer(asset.content);                                                             // 1262
    }                                                                                                        // 1263
    return asset;                                                                                            // 1264
  }                                                                                                          // 1265
                                                                                                             // 1266
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                                // 1267
    check(purpose, {profilePicture: {userId: DatabaseId, identityId: DatabaseId}});                          // 1268
                                                                                                             // 1269
    return AssetUploadTokens.insert({                                                                        // 1270
      purpose: purpose,                                                                                      // 1271
      expires: new Date(Date.now() + 300000),  // in 5 minutes                                               // 1272
    });                                                                                                      // 1273
  }                                                                                                          // 1274
                                                                                                             // 1275
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                                 // 1276
    // Indicates that the given asset upload has completed. It will be removed and its purpose               // 1277
    // returned. If no matching upload exists, returns undefined.                                            // 1278
                                                                                                             // 1279
    check(id, String);                                                                                       // 1280
                                                                                                             // 1281
    var upload = AssetUploadTokens.findAndModify({                                                           // 1282
      query: {_id: id},                                                                                      // 1283
      remove: true                                                                                           // 1284
    });                                                                                                      // 1285
                                                                                                             // 1286
    if (upload.expires.valueOf() < Date.now()) {                                                             // 1287
      return undefined;  // already expired                                                                  // 1288
    } else {                                                                                                 // 1289
      return upload.purpose;                                                                                 // 1290
    }                                                                                                        // 1291
  }                                                                                                          // 1292
                                                                                                             // 1293
  function cleanupExpiredAssetUploads() {                                                                    // 1294
    AssetUploadTokens.remove({expires: {$lt: Date.now()}});                                                  // 1295
  }                                                                                                          // 1296
                                                                                                             // 1297
  // Cleanup tokens every hour.                                                                              // 1298
  SandstormDb.periodicCleanup(3600000, cleanupExpiredAssetUploads);                                          // 1299
                                                                                                             // 1300
  var packageCache = {};                                                                                     // 1301
  // Package info is immutable. Let's cache to save on mongo queries.                                        // 1302
                                                                                                             // 1303
  SandstormDb.prototype.getPackage = function (packageId) {                                                  // 1304
    // Get the given package record. Since package info is immutable, cache the data in the server           // 1305
    // to reduce mongo query overhead, since it turns out we have to fetch specific packages a               // 1306
    // lot.                                                                                                  // 1307
                                                                                                             // 1308
    if (packageId in packageCache) {                                                                         // 1309
      return packageCache[packageId];                                                                        // 1310
    }                                                                                                        // 1311
                                                                                                             // 1312
    var pkg = Packages.findOne(packageId);                                                                   // 1313
    if (pkg && pkg.status === "ready") {                                                                     // 1314
      packageCache[packageId] = pkg;                                                                         // 1315
    }                                                                                                        // 1316
    return pkg;                                                                                              // 1317
  }                                                                                                          // 1318
                                                                                                             // 1319
  SandstormDb.prototype.sendAppUpdateNotifications = function (appId, packageId, name,                       // 1320
                                                               versionNumber, marketingVersion) {            // 1321
    var db = this;                                                                                           // 1322
    var actions = db.collections.userActions.find({appId: appId, appVersion: {$lt: versionNumber}},          // 1323
      {fields: {userId: 1}});                                                                                // 1324
    actions.forEach(function (action) {                                                                      // 1325
      var userId = action.userId;                                                                            // 1326
      var updater = {                                                                                        // 1327
        userId: userId,                                                                                      // 1328
        timestamp: new Date(),                                                                               // 1329
        isUnread: true,                                                                                      // 1330
      };                                                                                                     // 1331
                                                                                                             // 1332
      // Set only the appId that we care about. Use mongo's dot notation to specify only a single            // 1333
      // field inside of an object to update                                                                 // 1334
      updater["appUpdates." + appId] = {                                                                     // 1335
        marketingVersion: marketingVersion,                                                                  // 1336
        packageId: packageId,                                                                                // 1337
        name: name,                                                                                          // 1338
        version: versionNumber,                                                                              // 1339
      };                                                                                                     // 1340
      db.collections.notifications.upsert({userId: userId}, { $set: updater});                               // 1341
    });                                                                                                      // 1342
                                                                                                             // 1343
    db.collections.appIndex.update({_id: appId}, {$set: {hasSentNotifications: true}});                      // 1344
                                                                                                             // 1345
    // In the case where we replaced a previous notification and that was the only reference to the          // 1346
    // package, we need to clean it up                                                                       // 1347
    Meteor.call("deleteUnusedPackages", appId);                                                              // 1348
  };                                                                                                         // 1349
                                                                                                             // 1350
  SandstormDb.prototype.upgradeGrains =  function (appId, version, packageId, backend) {                     // 1351
    check(appId, String);                                                                                    // 1352
    check(version, Match.Integer);                                                                           // 1353
    check(packageId, String);                                                                                // 1354
                                                                                                             // 1355
    var selector = {                                                                                         // 1356
      userId: Meteor.userId(),                                                                               // 1357
      appId: appId,                                                                                          // 1358
      appVersion: { $lte: version },                                                                         // 1359
      packageId: { $ne: packageId }                                                                          // 1360
    };                                                                                                       // 1361
                                                                                                             // 1362
    if (!this.isSimulation) {                                                                                // 1363
      Grains.find(selector).forEach(function (grain) {                                                       // 1364
        backend.shutdownGrain(grain._id, grain.userId);                                                      // 1365
      });                                                                                                    // 1366
    }                                                                                                        // 1367
                                                                                                             // 1368
    Grains.update(selector, { $set: { appVersion: version, packageId: packageId }}, {multi: true});          // 1369
  };                                                                                                         // 1370
                                                                                                             // 1371
  SandstormDb.prototype.startInstall = function (packageId, url, retryFailed, isAutoUpdated) {               // 1372
    // Mark package for possible installation.                                                               // 1373
                                                                                                             // 1374
    var fields = {status: "download", progress: 0, url: url, isAutoUpdated: !!isAutoUpdated};                // 1375
                                                                                                             // 1376
    if (retryFailed) {                                                                                       // 1377
      Packages.update({_id: packageId, status: "failed"}, {$set: fields});                                   // 1378
    } else {                                                                                                 // 1379
      try {                                                                                                  // 1380
        fields._id = packageId;                                                                              // 1381
        Packages.insert(fields);                                                                             // 1382
      } catch (err) {                                                                                        // 1383
        console.error("Simultaneous startInstall()s?", err.stack);                                           // 1384
      }                                                                                                      // 1385
    }                                                                                                        // 1386
  };                                                                                                         // 1387
                                                                                                             // 1388
  var ValidKeyFingerprint = Match.Where(function (keyFingerprint) {                                          // 1389
    check(keyFingerprint, String);                                                                           // 1390
    return !!keyFingerprint.match(/^[0-9A-F]{40}$/);                                                         // 1391
  });                                                                                                        // 1392
                                                                                                             // 1393
  SandstormDb.prototype.updateKeybaseProfileAsync = function (keyFingerprint) {                              // 1394
    // Asynchronously fetch the given Keybase profile and populate the KeybaseProfiles collection.           // 1395
                                                                                                             // 1396
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1397
                                                                                                             // 1398
    console.log("fetching keybase", keyFingerprint);                                                         // 1399
                                                                                                             // 1400
    HTTP.get(                                                                                                // 1401
        "https://keybase.io/_/api/1.0/user/lookup.json?key_fingerprint=" + keyFingerprint +                  // 1402
        "&fields=basics,profile,proofs_summary", {                                                           // 1403
      timeout: 5000                                                                                          // 1404
    }, function (err, keybaseResponse) {                                                                     // 1405
      if (err) {                                                                                             // 1406
        console.log("keybase lookup error:", err.stack);                                                     // 1407
        return;                                                                                              // 1408
      }                                                                                                      // 1409
                                                                                                             // 1410
      if (!keybaseResponse.data) {                                                                           // 1411
        console.log("keybase didn't return JSON? Headers:", keybaseResponse.headers);                        // 1412
        return;                                                                                              // 1413
      }                                                                                                      // 1414
                                                                                                             // 1415
      var profile = (keybaseResponse.data.them || [])[0];                                                    // 1416
                                                                                                             // 1417
      if (profile) {                                                                                         // 1418
        var record = {                                                                                       // 1419
          displayName: (profile.profile || {}).full_name,                                                    // 1420
          handle: (profile.basics || {}).username,                                                           // 1421
          proofs: (profile.proofs_summary || {}).all || []                                                   // 1422
        };                                                                                                   // 1423
                                                                                                             // 1424
        record.proofs.forEach(function (proof) {                                                             // 1425
          // Remove potentially Mongo-incompatible stuff. (Currently Keybase returns nothing that            // 1426
          // this would filter.)                                                                             // 1427
          for (var field in proof) {                                                                         // 1428
            // Don't allow field names containing '.' or '$'. Also don't allow sub-objects mainly            // 1429
            // because I'm too lazy to check the field names recursively (and Keybase doesn't                // 1430
            // return any objects anyway).                                                                   // 1431
            if (field.match(/[.$]/) || typeof(proof[field]) === "object") {                                  // 1432
              delete proof[field];                                                                           // 1433
            }                                                                                                // 1434
          }                                                                                                  // 1435
                                                                                                             // 1436
          // Indicate not verified.                                                                          // 1437
          // TODO(security): Asynchronously verify proofs. Presumably we can borrow code from the            // 1438
          //   Keybase node-based CLI.                                                                       // 1439
          proof.status = "unverified";                                                                       // 1440
        });                                                                                                  // 1441
                                                                                                             // 1442
        KeybaseProfiles.update(keyFingerprint, {$set: record}, {upsert: true});                              // 1443
      } else {                                                                                               // 1444
        // Keybase reports no match, so remove what we know of this user. We don't want to remove            // 1445
        // the item entirely from the cache as this will cause us to repeatedly re-fetch the data            // 1446
        // from Keybase.                                                                                     // 1447
        //                                                                                                   // 1448
        // TODO(someday): We could perhaps keep the proofs if we can still verify them directly,             // 1449
        //   but at present we don't have the ability to verify proofs.                                      // 1450
        KeybaseProfiles.update(keyFingerprint,                                                               // 1451
            {$unset: {displayName: "", handle: "", proofs: ""}}, {upsert: true});                            // 1452
      }                                                                                                      // 1453
    });                                                                                                      // 1454
  };                                                                                                         // 1455
                                                                                                             // 1456
  Meteor.publish("keybaseProfile", function (keyFingerprint) {                                               // 1457
    check(keyFingerprint, ValidKeyFingerprint);                                                              // 1458
    var db = this.connection.sandstormDb;                                                                    // 1459
                                                                                                             // 1460
    var cursor = db.collections.keybaseProfiles.find(keyFingerprint);                                        // 1461
    if (cursor.count() === 0) {                                                                              // 1462
      // Fire off async update.                                                                              // 1463
      db.updateKeybaseProfileAsync(keyFingerprint);                                                          // 1464
    }                                                                                                        // 1465
                                                                                                             // 1466
    return cursor;                                                                                           // 1467
  });                                                                                                        // 1468
                                                                                                             // 1469
  Meteor.publish("appIndex", function (appId) {                                                              // 1470
    check(appId, String);                                                                                    // 1471
    var db = this.connection.sandstormDb;                                                                    // 1472
    var cursor = db.collections.appIndex.find({_id: appId});                                                 // 1473
    return cursor;                                                                                           // 1474
  });                                                                                                        // 1475
                                                                                                             // 1476
  Meteor.publish("userPackages", function() {                                                                // 1477
    // Users should be able to see packages that are either:                                                 // 1478
    // 1. referenced by one of their userActions                                                             // 1479
    // 2. referenced by one of their grains                                                                  // 1480
    var self = this;                                                                                         // 1481
    var db = this.connection.sandstormDb;                                                                    // 1482
                                                                                                             // 1483
    // Note that package information, once it is in the database, is static. There's no need to              // 1484
    // reactively subscribe to changes to a package since they don't change. It's also unecessary            // 1485
    // to reactively remove a package from the client side when it is removed on the server, or              // 1486
    // when the client stops using it, because the worst case is the client has a small amount               // 1487
    // of extra info on a no-longer-used package held in memory until they refresh Sandstorm.                // 1488
    // So, we implement this as a cache: the first time each package ID shows up among the user's            // 1489
    // stuff, we push the package info to the client, and then we never update it.                           // 1490
    //                                                                                                       // 1491
    // Alternatively, we could subscribe to each individual package query, but this would waste              // 1492
    // lots of server-side resources watching for events that will never happen or don't matter.             // 1493
    var hasPackage = {};                                                                                     // 1494
    var refPackage = function (packageId) {                                                                  // 1495
      // Ignore dev apps.                                                                                    // 1496
      if (packageId.lastIndexOf("dev-", 0) === 0) return;                                                    // 1497
                                                                                                             // 1498
      if (!hasPackage[packageId]) {                                                                          // 1499
        hasPackage[packageId] = true;                                                                        // 1500
        var pkg = db.getPackage(packageId);                                                                  // 1501
        if (pkg) {                                                                                           // 1502
          self.added("packages", packageId, pkg);                                                            // 1503
        } else {                                                                                             // 1504
          console.error(                                                                                     // 1505
              "shouldn't happen: missing package referenced by user's stuff:", packageId);                   // 1506
        }                                                                                                    // 1507
      }                                                                                                      // 1508
    };                                                                                                       // 1509
                                                                                                             // 1510
    // package source 1: packages referred to by actions                                                     // 1511
    var actions = db.userActions(this.userId);                                                               // 1512
    var actionsHandle = actions.observe({                                                                    // 1513
      added: function(newAction) {                                                                           // 1514
        refPackage(newAction.packageId);                                                                     // 1515
      },                                                                                                     // 1516
      changed: function(oldAction, newAction) {                                                              // 1517
        refPackage(newAction.packageId);                                                                     // 1518
      }                                                                                                      // 1519
    });                                                                                                      // 1520
                                                                                                             // 1521
    // package source 2: packages referred to by grains directly                                             // 1522
    var grains = db.userGrains(this.userId);                                                                 // 1523
    var grainsHandle = grains.observe({                                                                      // 1524
      added: function(newGrain) {                                                                            // 1525
        refPackage(newGrain.packageId);                                                                      // 1526
      },                                                                                                     // 1527
      changed: function(oldGrain, newGrain) {                                                                // 1528
        refPackage(newGrain.packageId);                                                                      // 1529
      }                                                                                                      // 1530
    });                                                                                                      // 1531
                                                                                                             // 1532
    this.onStop(function () {                                                                                // 1533
      actionsHandle.stop();                                                                                  // 1534
      grainsHandle.stop();                                                                                   // 1535
    });                                                                                                      // 1536
    this.ready();                                                                                            // 1537
  });                                                                                                        // 1538
}                                                                                                            // 1539
                                                                                                             // 1540
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/profile.js                                                                          //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var makeIdenticon;                                                                                           // 17
var httpProtocol;                                                                                            // 18
                                                                                                             // 19
if (Meteor.isServer) {                                                                                       // 20
  SandstormDb.ensureSubscriberHasIdentity = function(publishHandler, identityId) {                           // 21
    // Helper for publish functions that need to restrict access based on whether the subscriber             // 22
    // has a given identity linked. Automatically stops the subscription if the user loses the               // 23
    // identity. Returns a boolean indicating whether the user initially has the identity.                   // 24
                                                                                                             // 25
    var userId = publishHandler.userId;                                                                      // 26
    if (userId === identityId) {                                                                             // 27
      return true;                                                                                           // 28
    } else {                                                                                                 // 29
      var hasIdentityCursor =                                                                                // 30
          Meteor.users.find({$or: [{_id: userId, "loginIdentities.id": identityId},                          // 31
                                   {_id: userId, "nonloginIdentities.id": identityId}]});                    // 32
      if (hasIdentityCursor.count() == 0) {                                                                  // 33
        publishHandler.stop();                                                                               // 34
        return false;                                                                                        // 35
      }                                                                                                      // 36
      var handle = hasIdentityCursor.observe({removed: function () { publishHandler.stop(); }});             // 37
      publishHandler.onStop(function () { handle.stop(); });                                                 // 38
      return true;                                                                                           // 39
    }                                                                                                        // 40
  }                                                                                                          // 41
                                                                                                             // 42
  Meteor.publish("identityProfile", function (identityId) {                                                  // 43
    check(identityId, String);                                                                               // 44
    if (!SandstormDb.ensureSubscriberHasIdentity(this, identityId)) return;                                  // 45
                                                                                                             // 46
    return Meteor.users.find({_id: identityId},                                                              // 47
      {fields: {                                                                                             // 48
        "profile":1,                                                                                         // 49
        "unverifiedEmail":1,                                                                                 // 50
        "expires": 1,                                                                                        // 51
                                                                                                             // 52
        "services.dev.name":1,                                                                               // 53
                                                                                                             // 54
        "services.google.id":1,                                                                              // 55
        "services.google.email":1,                                                                           // 56
        "services.google.verified_email":1,                                                                  // 57
        "services.google.name":1,                                                                            // 58
        "services.google.picture":1,                                                                         // 59
        "services.google.gender":1,                                                                          // 60
                                                                                                             // 61
        "services.github.id":1,                                                                              // 62
        "services.github.email":1,                                                                           // 63
        "services.github.emails":1,                                                                          // 64
        "services.github.username":1,                                                                        // 65
                                                                                                             // 66
        "services.email.email":1,                                                                            // 67
      }});                                                                                                   // 68
  }),                                                                                                        // 69
                                                                                                             // 70
  Meteor.publish("accountIdentities", function () {                                                          // 71
    // Maybe this should be folded into the "credentials" subscription?                                      // 72
                                                                                                             // 73
    if (!this.userId) return [];                                                                             // 74
                                                                                                             // 75
    return Meteor.users.find(                                                                                // 76
      {_id: this.userId},                                                                                    // 77
      {fields: {                                                                                             // 78
        "profile": 1,                                                                                        // 79
        "verifiedEmail": 1,                                                                                  // 80
        "loginIdentities": 1,                                                                                // 81
        "nonloginIdentities": 1,                                                                             // 82
        "expires": 1,                                                                                        // 83
        "primaryEmail": 1,                                                                                   // 84
      }});                                                                                                   // 85
  });                                                                                                        // 86
                                                                                                             // 87
  makeIdenticon = function (id) {                                                                            // 88
    // We only make identicons client-side.                                                                  // 89
    return undefined;                                                                                        // 90
  }                                                                                                          // 91
                                                                                                             // 92
  var Url = Npm.require("url");                                                                              // 93
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                                   // 94
} else {                                                                                                     // 95
  var identiconCache = {};                                                                                   // 96
                                                                                                             // 97
  makeIdenticon = function (id) {                                                                            // 98
    // Given a cryptographic hash as input, generate an identicon. We always use the user's                  // 99
    // identity ID as the input hash (the same thing passed to apps as the user ID).                         // 100
                                                                                                             // 101
    // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so               // 102
    // that apps can themselves use identicon.js to produce consistent identicons. As it turns out           // 103
    // identicon.js doesn't use the second half of the hash even if we provide it, but slice it              // 104
    // anyway to be safe.                                                                                    // 105
    id = id.slice(0, 32);                                                                                    // 106
                                                                                                             // 107
    if (id in identiconCache) {                                                                              // 108
      return identiconCache[id];                                                                             // 109
    }                                                                                                        // 110
                                                                                                             // 111
    // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.             // 112
    var data = new Identicon(id, 512).toString();                                                            // 113
    var result = "data:image/png;base64," + data;                                                            // 114
    identiconCache[id] = result;                                                                             // 115
    return result;                                                                                           // 116
  }                                                                                                          // 117
                                                                                                             // 118
  httpProtocol = window.location.protocol;                                                                   // 119
}                                                                                                            // 120
                                                                                                             // 121
var GENDERS = {male: "male", female: "female", neutral: "neutral", robot: "robot"};                          // 122
                                                                                                             // 123
function staticAssetUrl(id, staticHost) {                                                                    // 124
  if (id) {                                                                                                  // 125
    return staticHost + "/" + id;                                                                            // 126
  } else {                                                                                                   // 127
    return undefined;                                                                                        // 128
  }                                                                                                          // 129
}                                                                                                            // 130
                                                                                                             // 131
function filterHandle(handle) {                                                                              // 132
  // Convert disallowed letters into underscores, but no more than one underscore in a row.                  // 133
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                                     // 134
}                                                                                                            // 135
                                                                                                             // 136
function emailToHandle(email) {                                                                              // 137
  // Turn an email address into a handle.                                                                    // 138
                                                                                                             // 139
  if (!email) return undefined;                                                                              // 140
                                                                                                             // 141
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters               // 142
  // on the same account.                                                                                    // 143
  var parts = email.split("@");                                                                              // 144
  var base = filterHandle(parts[0].split("+")[0]);                                                           // 145
                                                                                                             // 146
  var domain = (parts[1]||"").split(".");                                                                    // 147
  if (domain[domain.length - 1] === "name") {                                                                // 148
    // Oh, a .name domain. Let's use                                                                         // 149
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                       // 150
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info",                  // 151
                         "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"],                 // 152
                         base)) {                                                                            // 153
    // This is probably an address at a vanity domain. Use the domain itself as the handle.                  // 154
    base = filterHandle(domain[0]);                                                                          // 155
  }                                                                                                          // 156
                                                                                                             // 157
  return filterHandle(base);                                                                                 // 158
}                                                                                                            // 159
                                                                                                             // 160
SandstormDb.fillInProfileDefaults = function(user) {                                                         // 161
  var profile = user.profile;                                                                                // 162
  if (profile.service === "github") {                                                                        // 163
    profile.name = profile.name || user.services.github.username || "Name Unknown";                          // 164
    profile.handle = profile.handle || filterHandle(user.services.github.username) ||                        // 165
        filterHandle(profile.name);                                                                          // 166
  } else if (profile.service === "google") {                                                                 // 167
    profile.name = profile.name || user.services.google.name || "Name Unknown";                              // 168
    profile.handle = profile.handle || emailToHandle(user.services.google.email) ||                          // 169
        filterHandle(profile.name);                                                                          // 170
    profile.pronoun = profile.pronoun || GENDERS[user.services.google.gender] || "neutral";                  // 171
  } else if (profile.service === "email") {                                                                  // 172
    var email = user.services.email.email                                                                    // 173
    profile.name = profile.name || email.split("@")[0];                                                      // 174
    profile.handle = profile.handle || emailToHandle(email);                                                 // 175
  } else if (profile.service === "dev") {                                                                    // 176
    var lowerCaseName = user.services.dev.name.split(" ")[0].toLowerCase();                                  // 177
    profile.name = profile.name || user.services.dev.name;                                                   // 178
    profile.handle = profile.handle || filterHandle(lowerCaseName);                                          // 179
    profile.pronoun = profile.pronoun ||                                                                     // 180
        (_.contains(["alice", "carol", "eve"], lowerCaseName) ? "female" :                                   // 181
         _.contains(["bob", "dave"], lowerCaseName) ? "male" : "neutral");                                   // 182
  } else if (profile.service === "demo") {                                                                   // 183
    profile.name = profile.name || "Demo User";                                                              // 184
    profile.handle = profile.handle || "demo";                                                               // 185
  } else {                                                                                                   // 186
    throw new Error("unrecognized identity service: ", profile.service);                                     // 187
  }                                                                                                          // 188
                                                                                                             // 189
  profile.pronoun = profile.pronoun || "neutral";                                                            // 190
}                                                                                                            // 191
                                                                                                             // 192
SandstormDb.fillInIntrinsicName = function(user) {                                                           // 193
  var profile = user.profile;                                                                                // 194
  if (profile.service === "github") {                                                                        // 195
    profile.intrinsicName = user.services.github.username;                                                   // 196
  } else if (profile.service === "google") {                                                                 // 197
    profile.intrinsicName = user.services.google.name;                                                       // 198
    user.privateIntrinsicName = user.services.google.email;                                                  // 199
  } else if (profile.service === "email") {                                                                  // 200
    profile.intrinsicName = user.services.email.email                                                        // 201
  } else if (profile.service === "dev") {                                                                    // 202
    profile.intrinsicName = user.services.dev.name;                                                          // 203
  } else if (profile.service === "demo") {                                                                   // 204
    // No intrinsic name.                                                                                    // 205
  } else {                                                                                                   // 206
    throw new Error("unrecognized identity service: ", profile.service);                                     // 207
  }                                                                                                          // 208
}                                                                                                            // 209
                                                                                                             // 210
SandstormDb.getVerifiedEmails = function(identity) {                                                         // 211
  if (identity.services.google && identity.services.google.email &&                                          // 212
      identity.services.google.verified_email) {                                                             // 213
    return [identity.services.google.email];                                                                 // 214
  } else if (identity.services.email) {                                                                      // 215
    return [identity.services.email.email];                                                                  // 216
  } else if (identity.services.github && identity.services.github.emails) {                                  // 217
    return _.pluck(_.filter(identity.services.github.emails,                                                 // 218
                            function(email) { return email.verified; }),                                     // 219
                   "email");                                                                                 // 220
  }                                                                                                          // 221
  return [];                                                                                                 // 222
}                                                                                                            // 223
                                                                                                             // 224
SandstormDb.fillInPictureUrl = function(user) {                                                              // 225
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                         // 226
  user.profile.pictureUrl = staticAssetUrl(user.profile.picture, staticHost) ||                              // 227
    makeIdenticon(user._id);                                                                                 // 228
}                                                                                                            // 229
                                                                                                             // 230
SandstormDb.getUserIdentityIds = function (user) {                                                           // 231
  // Given an account user object, returns an array containing the ID of each identity linked to the         // 232
  // account. Always returns the most recently added login identity first.                                   // 233
  if (user && user.loginIdentities) {                                                                        // 234
    return _.pluck(user.nonloginIdentities.concat(user.loginIdentities), "id").reverse();                    // 235
  } else {                                                                                                   // 236
    return [];                                                                                               // 237
  }                                                                                                          // 238
};                                                                                                           // 239
                                                                                                             // 240
SandstormDb.getUserEmails = function (user) {                                                                // 241
  // Given a user object, returns an array containing all email addresses associated with that user.         // 242
  // Each entry in the array is an object of the form:                                                       // 243
  //     `{email: String, verified: Bool, primary: Optional(Bool)}`                                          // 244
  //                                                                                                         // 245
  // At most one entry in the result has `primary = true`.                                                   // 246
  //                                                                                                         // 247
  // TODO(cleanup): This actually does need to query the database to fetch profile information               // 248
  //   for linked identities, so it probably makes more sense for it to be a non-static method               // 249
  //   on SandstormDb.                                                                                       // 250
                                                                                                             // 251
  var identityIds = SandstormDb.getUserIdentityIds(user);                                                    // 252
  verifiedEmails = {};                                                                                       // 253
  unverifiedEmails = {};                                                                                     // 254
                                                                                                             // 255
  identityIds.forEach(function (id) {                                                                        // 256
    var identity = Meteor.users.findOne({_id: id});                                                          // 257
    if (identity && identity.services) {                                                                     // 258
      SandstormDb.getVerifiedEmails(identity).forEach(function(verifiedEmail) {                              // 259
        if (verifiedEmail) {                                                                                 // 260
          verifiedEmails[verifiedEmail] = true;                                                              // 261
        }                                                                                                    // 262
      });                                                                                                    // 263
    }                                                                                                        // 264
    if (identity && identity.unverifiedEmail) {                                                              // 265
      unverifiedEmails[identity.unverifiedEmail] = true;                                                     // 266
    }                                                                                                        // 267
  });                                                                                                        // 268
                                                                                                             // 269
  var result = [];                                                                                           // 270
  _.keys(verifiedEmails).map(function (email) {                                                              // 271
    result.push({email: email,                                                                               // 272
                 verified: true,                                                                             // 273
                 primary: email === user.primaryEmail});                                                     // 274
  });                                                                                                        // 275
  _.keys(unverifiedEmails).map(function (email) {                                                            // 276
    if (!(email in verifiedEmails)) { result.push({email: email, verified: false}) }                         // 277
  });                                                                                                        // 278
                                                                                                             // 279
  // If `user.primaryEmail` is not among the verified emails, mark the first as primary.                     // 280
  if (!(user.primaryEmail in verifiedEmails) && result.length > 0) {                                         // 281
    result[0].primary = true;                                                                                // 282
  }                                                                                                          // 283
  return result;                                                                                             // 284
}                                                                                                            // 285
                                                                                                             // 286
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/user.js                                                                             //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Sandstorm - Personal Cloud Sandbox                                                                        // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                     // 2
// All rights reserved.                                                                                      // 3
//                                                                                                           // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                           // 5
// you may not use this file except in compliance with the License.                                          // 6
// You may obtain a copy of the License at                                                                   // 7
//                                                                                                           // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                              // 9
//                                                                                                           // 10
// Unless required by applicable law or agreed to in writing, software                                       // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                         // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                  // 13
// See the License for the specific language governing permissions and                                       // 14
// limitations under the License.                                                                            // 15
                                                                                                             // 16
var Crypto = Npm.require("crypto");                                                                          // 17
var Future = Npm.require("fibers/future");                                                                   // 18
                                                                                                             // 19
userPictureUrl = function (user) {                                                                           // 20
  if (user.services && !(user.profile && user.profile.picture)) {                                            // 21
    // Try to determine user's avatar URL from login service.                                                // 22
                                                                                                             // 23
    var google = user.services.google;                                                                       // 24
    if (google && google.picture) {                                                                          // 25
      return google.picture;                                                                                 // 26
    }                                                                                                        // 27
                                                                                                             // 28
    var github = user.services.github;                                                                       // 29
    if (github && github.id) {                                                                               // 30
      return "https://avatars.githubusercontent.com/u/" + github.id;                                         // 31
    }                                                                                                        // 32
                                                                                                             // 33
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be            // 34
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and                // 35
    // Github are different because they are actually the identity providers, so they already know           // 36
    // the user logged in.                                                                                   // 37
  }                                                                                                          // 38
}                                                                                                            // 39
                                                                                                             // 40
fetchPicture = function (url) {                                                                              // 41
  try {                                                                                                      // 42
    var result = HTTP.get(url, {                                                                             // 43
      npmRequestOptions: { encoding: null },                                                                 // 44
      timeout: 5000                                                                                          // 45
    });                                                                                                      // 46
                                                                                                             // 47
    var metadata = {};                                                                                       // 48
                                                                                                             // 49
    metadata.mimeType = result.headers["content-type"];                                                      // 50
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 &&                                              // 51
        metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {                                             // 52
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                        // 53
    }                                                                                                        // 54
    var enc = result.headers["content-encoding"];                                                            // 55
    if (enc && enc !== "identity") {                                                                         // 56
      metadata.encoding = enc;                                                                               // 57
    }                                                                                                        // 58
                                                                                                             // 59
    return addStaticAsset(metadata, result.content);                                                         // 60
  } catch (err) {                                                                                            // 61
    console.error("failed to fetch user profile picture:", url, err.stack);                                  // 62
  }                                                                                                          // 63
}                                                                                                            // 64
                                                                                                             // 65
var ValidHandle = Match.Where(function (handle) {                                                            // 66
  check(handle, String);                                                                                     // 67
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                               // 68
});                                                                                                          // 69
                                                                                                             // 70
Accounts.onCreateUser(function (options, user) {                                                             // 71
  if (user.loginIdentities) {                                                                                // 72
    // it's an account                                                                                       // 73
    check(user, {_id: String,                                                                                // 74
                 createdAt: Date,                                                                            // 75
                 isAdmin: Match.Optional(Boolean),                                                           // 76
                 hasCompletedSignup: Match.Optional(Boolean),                                                // 77
                 signupKey: Match.Optional(String),                                                          // 78
                 signupNote: Match.Optional(String),                                                         // 79
                 signupEmail: Match.Optional(String),                                                        // 80
                 expires: Match.Optional(Date),                                                              // 81
                 loginIdentities: [{id: String}],                                                            // 82
                 nonloginIdentities: [{id: String}]});                                                       // 83
    return user;                                                                                             // 84
  }                                                                                                          // 85
                                                                                                             // 86
  // Check profile.                                                                                          // 87
  if (options.profile) {                                                                                     // 88
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                                     // 89
    check(options.profile, Match.ObjectIncluding({                                                           // 90
      name: Match.OneOf(null, Match.Optional(String)),                                                       // 91
      handle: Match.Optional(ValidHandle),                                                                   // 92
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot")),                            // 93
    }));                                                                                                     // 94
  }                                                                                                          // 95
                                                                                                             // 96
  check(options.unverifiedEmail, Match.Optional(String));                                                    // 97
                                                                                                             // 98
  if (options.unverifiedEmail) {                                                                             // 99
    user.unverifiedEmail = options.unverifiedEmail;                                                          // 100
  }                                                                                                          // 101
  user.profile = _.pick(options.profile || {}, "name", "handle", "pronouns");                                // 102
                                                                                                             // 103
  // Try downloading avatar.                                                                                 // 104
  var url = userPictureUrl(user);                                                                            // 105
  if (url) {                                                                                                 // 106
    var assetId = fetchPicture(url);                                                                         // 107
    if (assetId) {                                                                                           // 108
      user.profile.picture = assetId;                                                                        // 109
    }                                                                                                        // 110
  }                                                                                                          // 111
                                                                                                             // 112
  var serviceUserId;                                                                                         // 113
  if (user.services && user.services.dev) {                                                                  // 114
    check(user.services.dev, {name: String, isAdmin: Boolean, hasCompletedSignup: Boolean});                 // 115
    serviceUserId = user.services.dev.name;                                                                  // 116
    user.profile.service = "dev";                                                                            // 117
  } else if ("expires" in user) {                                                                            // 118
    serviceUserId = user._id;                                                                                // 119
    user.profile.service = "demo";                                                                           // 120
  } else if (user.services && user.services.email) {                                                         // 121
    check(user.services.email,                                                                               // 122
          {email: String,                                                                                    // 123
           tokens: [{digest: String, algorithm: String, createdAt: Date}]});                                 // 124
    serviceUserId = user.services.email.email;                                                               // 125
    user.profile.service = "email";                                                                          // 126
  } else if (user.services && "google" in user.services) {                                                   // 127
    serviceUserId = user.services.google.id;                                                                 // 128
    user.profile.service = "google";                                                                         // 129
  } else if (user.services && "github" in user.services) {                                                   // 130
    serviceUserId = user.services.github.id;                                                                 // 131
    user.profile.service = "github";                                                                         // 132
  } else {                                                                                                   // 133
    throw new Meteor.Error(400, "user does not have a recognized identity provider: " +                      // 134
                           JSON.stringify(user));                                                            // 135
  }                                                                                                          // 136
  user._id = Crypto.createHash("sha256")                                                                     // 137
    .update(user.profile.service + ":" + serviceUserId).digest("hex");                                       // 138
                                                                                                             // 139
  return user;                                                                                               // 140
});                                                                                                          // 141
                                                                                                             // 142
// TODO delete this obsolete index.                                                                          // 143
Meteor.users._ensureIndex("identities.id", {unique: 1, sparse: 1});                                          // 144
                                                                                                             // 145
Meteor.users._ensureIndex("loginIdentities.id", {unique: 1, sparse: 1});                                     // 146
Meteor.users._ensureIndex("nonloginIdentities.id", {sparse: 1});                                             // 147
                                                                                                             // 148
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                           //
// packages/sandstorm-db/migrations.js                                                                       //
//                                                                                                           //
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                             //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                          // 1
//                                                                                                           // 2
// Because migrations can experience partial failure and likely have                                         // 3
// side-effects, we should be careful to make sure all migrations are                                        // 4
// idempotent and safe to accidentally run multiple times.                                                   // 5
                                                                                                             // 6
var Future = Npm.require("fibers/future");                                                                   // 7
                                                                                                             // 8
var updateLoginStyleToRedirect = function() {                                                                // 9
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 10
  ["google", "github"].forEach(function(serviceName) {                                                       // 11
    var config = configurations.findOne({service: serviceName});                                             // 12
    if (config && config.loginStyle !== "redirect") {                                                        // 13
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                       // 14
    }                                                                                                        // 15
  });                                                                                                        // 16
};                                                                                                           // 17
                                                                                                             // 18
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                                 // 19
  // In the before time, Google and Github login were enabled by default.                                    // 20
  //                                                                                                         // 21
  // This actually didn't make much sense, required the first user to configure                              // 22
  // OAuth, and had some trust-the-first-user properties that weren't totally                                // 23
  // secure.                                                                                                 // 24
  //                                                                                                         // 25
  // Now, we have admin-token, so we wish to disable all logins by default                                   // 26
  // (since they need to be configured anyway) but since users may have never                                // 27
  // explicitly told Sandstorm that Google or Github login should be enabled,                                // 28
  // we can't just use the value in the Settings collection, since it might                                  // 29
  // never have been set.                                                                                    // 30
  //                                                                                                         // 31
  // Thus, if the service is configured but absent from Settings, we should                                  // 32
  // explicitly enable it in Settings, and then the rest of the logic can just                               // 33
  // depend on what value is in Settings and default to false without breaking                               // 34
  // user installations.                                                                                     // 35
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                 // 36
  ["google", "github"].forEach(function(serviceName) {                                                       // 37
    var config = configurations.findOne({service: serviceName});                                             // 38
    var serviceConfig = Settings.findOne({_id: serviceName});                                                // 39
    if (config && !serviceConfig) {                                                                          // 40
      // Only explicitly enable the login service if:                                                        // 41
      // 1) the service is already configured                                                                // 42
      // 2) there is no sandstorm configuration already present (the user was                                // 43
      //    using the previous default behavior).                                                            // 44
      Settings.insert({_id: serviceName, value: true});                                                      // 45
    }                                                                                                        // 46
  });                                                                                                        // 47
};                                                                                                           // 48
                                                                                                             // 49
var denormalizeInviteInfo = function() {                                                                     // 50
  // When a user is invited via a signup token, the `signupKey` field of their user table entry              // 51
  // has always been populated to indicate the key they used. This points into the SignupKeys table          // 52
  // which has more information about the key, namely a freeform note entered by the admin when              // 53
  // they created the key. In the case that the email invite form was used, the note has the form            // 54
  // "E-mail invite to <address>".                                                                           // 55
  //                                                                                                         // 56
  // Later, we decided it was useful to indicate in the users table visible to the admin                     // 57
  // information about the invite terms. Namely, for email invites we want to show the address               // 58
  // and for others we want to show the note. To make this efficient, fields `signupNote` and                // 59
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing              // 60
  // from the SignupKeys table.                                                                              // 61
                                                                                                             // 62
  Meteor.users.find().forEach(function (user) {                                                              // 63
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {              // 64
      var signupInfo = SignupKeys.findOne(user.signupKey);                                                   // 65
      if (signupInfo && signupInfo.note) {                                                                   // 66
        var newFields = { signupNote: signupInfo.note };                                                     // 67
                                                                                                             // 68
        var prefix = "E-mail invite to ";                                                                    // 69
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                                     // 70
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                                      // 71
        }                                                                                                    // 72
                                                                                                             // 73
        Meteor.users.update(user._id, {$set: newFields});                                                    // 74
      }                                                                                                      // 75
    }                                                                                                        // 76
  });                                                                                                        // 77
}                                                                                                            // 78
                                                                                                             // 79
function mergeRoleAssignmentsIntoApiTokens() {                                                               // 80
  RoleAssignments.find().forEach(function(roleAssignment) {                                                  // 81
    ApiTokens.insert({                                                                                       // 82
      grainId: roleAssignment.grainId,                                                                       // 83
      userId: roleAssignment.sharer,                                                                         // 84
      roleAssignment: roleAssignment.roleAssignment,                                                         // 85
      petname: roleAssignment.petname,                                                                       // 86
      created: roleAssignment.created,                                                                       // 87
      owner: {user: {userId: roleAssignment.recipient,                                                       // 88
                     title: roleAssignment.title}},                                                          // 89
    });                                                                                                      // 90
  });                                                                                                        // 91
}                                                                                                            // 92
                                                                                                             // 93
function fixOasisStorageUsageStats() {}                                                                      // 94
// This migration only pertained to Oasis and it was successfully applied there. Since it referred           // 95
// to some global variables that we later wanted to remove and/or rename, we've since replaced it            // 96
// with a no-op.                                                                                             // 97
                                                                                                             // 98
function fetchProfilePictures() {                                                                            // 99
  Meteor.users.find({}).forEach(function (user) {                                                            // 100
    var url = userPictureUrl(user);                                                                          // 101
    if (url) {                                                                                               // 102
      console.log("Fetching user picture:", url);                                                            // 103
      var assetId = fetchPicture(url);                                                                       // 104
      if (assetId) {                                                                                         // 105
        Meteor.users.update(user._id, {$set: {"profile.picture": assetId}});                                 // 106
      }                                                                                                      // 107
    }                                                                                                        // 108
  });                                                                                                        // 109
}                                                                                                            // 110
                                                                                                             // 111
function assignPlans() {                                                                                     // 112
  if (Meteor.settings.public.quotaEnabled && SandstormDb.paymentsMigrationHook) {                            // 113
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                                     // 114
  }                                                                                                          // 115
}                                                                                                            // 116
                                                                                                             // 117
function removeKeyrings() {                                                                                  // 118
  // These blobs full of public keys were not intended to find their way into mongo and while                // 119
  // harmless they slow things down because they're huge. Remove them.                                       // 120
  Packages.update({"manifest.metadata.pgpKeyring": {$exists: true}},                                         // 121
      {$unset: {"manifest.metadata.pgpKeyring": ""}},                                                        // 122
      {multi: true});                                                                                        // 123
}                                                                                                            // 124
                                                                                                             // 125
function useLocalizedTextInUserActions() {                                                                   // 126
  function toLocalizedText(newObj, oldObj, field) {                                                          // 127
    if (field in oldObj) {                                                                                   // 128
      if (typeof oldObj[field] === "string") {                                                               // 129
        newObj[field] = {defaultText: oldObj[field]};                                                        // 130
      } else {                                                                                               // 131
        newObj[field] = oldObj[field];                                                                       // 132
      }                                                                                                      // 133
    }                                                                                                        // 134
  }                                                                                                          // 135
  UserActions.find({}).forEach(function (userAction) {                                                       // 136
    var fields = {};                                                                                         // 137
    toLocalizedText(fields, userAction, "appTitle");                                                         // 138
    toLocalizedText(fields, userAction, "title");                                                            // 139
    toLocalizedText(fields, userAction, "nounPhrase");                                                       // 140
    UserActions.update(userAction._id, {$set: fields});                                                      // 141
  });                                                                                                        // 142
}                                                                                                            // 143
                                                                                                             // 144
function verifyAllPgpSignatures() {                                                                          // 145
  Packages.find({}).forEach(function (pkg) {                                                                 // 146
    try {                                                                                                    // 147
      console.log("checking PGP signature for package:", pkg._id);                                           // 148
      var info = waitPromise(globalBackend.cap().tryGetPackage(pkg._id));                                    // 149
      if (info.authorPgpKeyFingerprint) {                                                                    // 150
        console.log("  " + info.authorPgpKeyFingerprint);                                                    // 151
        Packages.update(pkg._id,                                                                             // 152
            {$set: {authorPgpKeyFingerprint: info.authorPgpKeyFingerprint}});                                // 153
      } else {                                                                                               // 154
        console.log("  no signature");                                                                       // 155
      }                                                                                                      // 156
    } catch (err) {                                                                                          // 157
      console.error(err.stack);                                                                              // 158
    }                                                                                                        // 159
  });                                                                                                        // 160
}                                                                                                            // 161
                                                                                                             // 162
function splitUserIdsIntoAccountIdsAndIdentityIds() {                                                        // 163
  var Crypto = Npm.require("crypto");                                                                        // 164
  Meteor.users.find().forEach(function (user) {                                                              // 165
    var identity = {};                                                                                       // 166
    var serviceUserId;                                                                                       // 167
    if ("devName" in user) {                                                                                 // 168
      identity.service = "dev";                                                                              // 169
      serviceUserId = user.devName;                                                                          // 170
    } else if ("expires" in user) {                                                                          // 171
      identity.service = "demo";                                                                             // 172
      serviceUserId = user._id;                                                                              // 173
    } else if (user.services && "google" in user.services) {                                                 // 174
      identity.service = "google";                                                                           // 175
      if (user.services.google.email && user.services.google.verified_email) {                               // 176
        identity.verifiedEmail = user.services.google.email;                                                 // 177
      }                                                                                                      // 178
      serviceUserId = user.services.google.id;                                                               // 179
    } else if (user.services && "github" in user.services) {                                                 // 180
      identity.service = "github";                                                                           // 181
      identity.unverifiedEmail = user.services.github.email;                                                 // 182
      serviceUserId = user.services.github.id;                                                               // 183
    } else if (user.services && "emailToken" in user.services) {                                             // 184
      identity.service = "emailToken";                                                                       // 185
      identity.verifiedEmail = user.services.emailToken.email;                                               // 186
      serviceUserId = user.services.emailToken.email;                                                        // 187
    }                                                                                                        // 188
                                                                                                             // 189
    identity.id = Crypto.createHash("sha256")                                                                // 190
        .update(identity.service + ":" + serviceUserId).digest("hex");                                       // 191
                                                                                                             // 192
    if (user.profile) {                                                                                      // 193
      if (user.profile.name) {                                                                               // 194
        identity.name = user.profile.name;                                                                   // 195
      }                                                                                                      // 196
      if (user.profile.handle) {                                                                             // 197
        identity.handle = user.profile.handle;                                                               // 198
      }                                                                                                      // 199
      if (user.profile.picture) {                                                                            // 200
        identity.picture = user.profile.picture;                                                             // 201
      }                                                                                                      // 202
      if (user.profile.pronoun) {                                                                            // 203
        identity.pronoun = user.profile.pronoun;                                                             // 204
      }                                                                                                      // 205
      if (user.profile.email) {                                                                              // 206
        identity.unverifiedEmail = user.profile.email;                                                       // 207
      }                                                                                                      // 208
    }                                                                                                        // 209
    identity.main = true;                                                                                    // 210
                                                                                                             // 211
    Meteor.users.update(user._id, {$set: {identities: [identity]}});                                         // 212
                                                                                                             // 213
    Grains.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                     // 214
    Sessions.update({userId: user._id}, {$set: {identityId: identity.id}}, {multi: true});                   // 215
    ApiTokens.update({userId: user._id},                                                                     // 216
                     {$set: {identityId: identity.id}},                                                      // 217
                     {multi: true});                                                                         // 218
    ApiTokens.update({"owner.user.userId": user._id},                                                        // 219
                     {$set: {"owner.user.identityId": identity.id}},                                         // 220
                     {multi: true});                                                                         // 221
    ApiTokens.update({"owner.grain.introducerUser": user._id},                                               // 222
                     {$set: {"owner.grain.introducerIdentity": identity.id}},                                // 223
                     {multi: true});                                                                         // 224
                                                                                                             // 225
    while (ApiTokens.update({"requirements.permissionsHeld.userId": user._id},                               // 226
                            {$set: {"requirements.$.permissionsHeld.identityId": identity.id},               // 227
                             $unset: {"requirements.$.permissionsHeld.userId": 1}},                          // 228
                            {multi: true}) > 0);                                                             // 229
    // The `$` operatorer modifies the first element in the array that matches the query. Since              // 230
    // there may be many matches, we need to repeat until no documents are modified.                         // 231
                                                                                                             // 232
  });                                                                                                        // 233
                                                                                                             // 234
  ApiTokens.remove({userInfo: {$exists: true}});                                                             // 235
  // We've renamed `Grain.UserInfo.userId` to `Grain.userInfo.identityId`. The only place                    // 236
  // that this field could show up in the database was in this deprecated, no-longer-functional              // 237
  // form of API token.                                                                                      // 238
}                                                                                                            // 239
                                                                                                             // 240
function appUpdateSettings() {                                                                               // 241
  Settings.insert({_id: "appMarketUrl", value: "https://apps.sandstorm.io"});                                // 242
  Settings.insert({_id: "appIndexUrl", value: "https://app-index.sandstorm.io"});                            // 243
  Settings.insert({_id: "appUpdatesEnabled", value: true});                                                  // 244
}                                                                                                            // 245
                                                                                                             // 246
function moveDevAndEmailLoginDataIntoIdentities() {                                                          // 247
  Meteor.users.find().forEach(function (user) {                                                              // 248
    if (user.identities.length != 1) {                                                                       // 249
      throw new Error("User does not have exactly one identity: ", user);                                    // 250
    }                                                                                                        // 251
    var identity = user.identities[0];                                                                       // 252
    if (Match.test(identity.service, Object)) { return; } // Already migrated.                               // 253
                                                                                                             // 254
    var newIdentity = _.pick(identity, "id", "main", "noLogin", "verifiedEmail", "unverifiedEmail");         // 255
    newIdentity.profile = _.pick(identity, "name", "handle", "picture", "pronoun");                          // 256
                                                                                                             // 257
    var serviceObject = {};                                                                                  // 258
    var fieldsToUnset = {}                                                                                   // 259
                                                                                                             // 260
    if (identity.service === "dev") {                                                                        // 261
      serviceObject.name = user.devName;                                                                     // 262
      fieldsToUnset.devName = 1;                                                                             // 263
    } else if (identity.service === "emailToken") {                                                          // 264
      serviceObject.tokens = user.services.emailToken.tokens;                                                // 265
      serviceObject.email = user.services.emailToken.email;                                                  // 266
      fieldsToUnset["services.emailToken"] = 1;                                                              // 267
    }                                                                                                        // 268
                                                                                                             // 269
    newIdentity.service = {};                                                                                // 270
    newIdentity.service[identity.service] = serviceObject;                                                   // 271
                                                                                                             // 272
    var modifier = {$set: {identities: [newIdentity]}};                                                      // 273
    if (Object.keys(fieldsToUnset).length > 0) {                                                             // 274
      modifier["$unset"] = fieldsToUnset;                                                                    // 275
    }                                                                                                        // 276
                                                                                                             // 277
    Meteor.users.update({_id: user._id}, modifier);                                                          // 278
  });                                                                                                        // 279
}                                                                                                            // 280
                                                                                                             // 281
function repairEmailIdentityIds() {                                                                          // 282
  var Crypto = Npm.require("crypto");                                                                        // 283
  Meteor.users.find({"identities.service.emailToken": {$exists: 1}}).forEach(function (user) {               // 284
    if (user.identities.length != 1) {                                                                       // 285
      throw new Error("User does not have exactly one identity: ", user);                                    // 286
    }                                                                                                        // 287
    var identity = user.identities[0];                                                                       // 288
    var newIdentity = _.pick(identity, "main", "noLogin", "verifiedEmail", "unverifiedMail",                 // 289
                             "profile");                                                                     // 290
    newIdentity.service = {email: identity.service.emailToken};                                              // 291
    newIdentity.id = Crypto.createHash("sha256")                                                             // 292
      .update("email:" + identity.service.emailToken.email).digest("hex");                                   // 293
                                                                                                             // 294
    Grains.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});           // 295
    Sessions.update({identityId: identity.id}, {$set: {identityId: newIdentity.id}}, {multi: true});         // 296
    ApiTokens.update({identityId: identity.id},                                                              // 297
                     {$set: {identityId: newIdentity.id}},                                                   // 298
                     {multi: true});                                                                         // 299
    ApiTokens.update({"owner.user.identityId": identity.id},                                                 // 300
                     {$set: {"owner.user.identityId": newIdentity.id}},                                      // 301
                     {multi: true});                                                                         // 302
    ApiTokens.update({"owner.grain.introducerIdentity": identity.id},                                        // 303
                     {$set: {"owner.grain.introducerIdentity": newIdentity.id}},                             // 304
                     {multi: true});                                                                         // 305
                                                                                                             // 306
    while (ApiTokens.update({"requirements.permissionsHeld.identityId": identity.id},                        // 307
                            {$set: {"requirements.$.permissionsHeld.identityId": newIdentity.id}},           // 308
                            {multi: true}) > 0);                                                             // 309
                                                                                                             // 310
    Meteor.users.update({_id: user._id}, {$set: {identities: [newIdentity]}});                               // 311
  });                                                                                                        // 312
}                                                                                                            // 313
                                                                                                             // 314
function splitAccountUsersAndIdentityUsers() {                                                               // 315
  Meteor.users.find({"identities": {$exists: true}}).forEach(function (user) {                               // 316
    if (user.identities.length != 1) {                                                                       // 317
      throw new Error("User does not have exactly one identity: ", user);                                    // 318
    }                                                                                                        // 319
    var identity = user.identities[0];                                                                       // 320
    var identityUser = _.pick(user, "createdAt", "lastActive", "expires")                                    // 321
    identityUser._id = identity.id;                                                                          // 322
    identityUser.profile = identity.profile;                                                                 // 323
    _.extend(identityUser, _.pick(identity, "unverifiedEmail"));                                             // 324
    identityUser.profile.service = Object.keys(identity.service)[0];                                         // 325
                                                                                                             // 326
    // Updating this user needs to be a two step process because the `services` field typically              // 327
    // contains subfields that are constrained to be unique by Mongo indices.                                // 328
    identityUser.stagedServices = _.omit(user.services, "resume");                                           // 329
    if (identity.service.dev) {                                                                              // 330
      identityUser.stagedServices.dev = identity.service.dev;                                                // 331
    } else if (identity.service.email) {                                                                     // 332
      identityUser.stagedServices.email = identity.service.email;                                            // 333
    }                                                                                                        // 334
                                                                                                             // 335
    var accountUser = _.pick(user, "_id", "createdAt", "lastActive", "expires",                              // 336
                             "isAdmin", "signupKey", "signupNote", "signupEmail",                            // 337
                             "plan", "storageUsage", "isAppDemoUser", "appDemoId",                           // 338
                             "payments", "dailySentMailCount", "hasCompletedSignup");                        // 339
    accountUser.loginIdentities = [_.pick(identity, "id")];                                                  // 340
    accountUser.nonloginIdentities = [];                                                                     // 341
    if (user.services && user.services.resume) {                                                             // 342
      accountUser.services = {resume: user.services.resume};                                                 // 343
    }                                                                                                        // 344
                                                                                                             // 345
    accountUser.stashedOldUser = user;                                                                       // 346
                                                                                                             // 347
    ApiTokens.update({identityId: identityUser._id}, {$set: {accountId: user._id}},                          // 348
                     {multi: true});                                                                         // 349
                                                                                                             // 350
    Meteor.users.upsert({_id: identityUser._id}, identityUser);                                              // 351
    Meteor.users.update({_id: user._id}, accountUser);                                                       // 352
  });                                                                                                        // 353
                                                                                                             // 354
  Meteor.users.find({"stagedServices": {$exists: true}}).forEach(function (identity) {                       // 355
    Meteor.users.update({_id: identity._id}, {$unset: {stagedServices: 1},                                   // 356
                                              $set: {services: identity.stagedServices}});                   // 357
  });                                                                                                        // 358
}                                                                                                            // 359
                                                                                                             // 360
// This must come after all the functions named within are defined.                                          // 361
// Only append to this list!  Do not modify or remove list entries;                                          // 362
// doing so is likely change the meaning and semantics of user databases.                                    // 363
var MIGRATIONS = [                                                                                           // 364
  updateLoginStyleToRedirect,                                                                                // 365
  enableLegacyOAuthProvidersIfNotInSettings,                                                                 // 366
  denormalizeInviteInfo,                                                                                     // 367
  mergeRoleAssignmentsIntoApiTokens,                                                                         // 368
  fixOasisStorageUsageStats,                                                                                 // 369
  fetchProfilePictures,                                                                                      // 370
  assignPlans,                                                                                               // 371
  removeKeyrings,                                                                                            // 372
  useLocalizedTextInUserActions,                                                                             // 373
  verifyAllPgpSignatures,                                                                                    // 374
  splitUserIdsIntoAccountIdsAndIdentityIds,                                                                  // 375
  appUpdateSettings,                                                                                         // 376
  moveDevAndEmailLoginDataIntoIdentities,                                                                    // 377
  repairEmailIdentityIds,                                                                                    // 378
  splitAccountUsersAndIdentityUsers,                                                                         // 379
];                                                                                                           // 380
                                                                                                             // 381
function migrateToLatest() {                                                                                 // 382
  if (Meteor.settings.replicaNumber) {                                                                       // 383
    // This is a replica. Wait for the first replica to perform migrations.                                  // 384
                                                                                                             // 385
    console.log("Waiting for migrations on replica zero...");                                                // 386
                                                                                                             // 387
    var done = new Future();                                                                                 // 388
    var change = function (doc) {                                                                            // 389
      console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);                   // 390
      if (doc.value >= MIGRATIONS.length) done.return();                                                     // 391
    }                                                                                                        // 392
    var observer = Migrations.find({_id: "migrations_applied"}).observe({                                    // 393
      added: change,                                                                                         // 394
      changed: change                                                                                        // 395
    });                                                                                                      // 396
                                                                                                             // 397
    done.wait();                                                                                             // 398
    observer.stop();                                                                                         // 399
    console.log("Migrations have completed on replica zero.");                                               // 400
                                                                                                             // 401
  } else {                                                                                                   // 402
    var applied = Migrations.findOne({_id: "migrations_applied"});                                           // 403
    var start;                                                                                               // 404
    if (!applied) {                                                                                          // 405
      // Migrations table is not yet seeded with a value.  This means it has                                 // 406
      // applied 0 migrations.  Persist this.                                                                // 407
      Migrations.insert({_id: "migrations_applied", value: 0});                                              // 408
      start = 0;                                                                                             // 409
    } else {                                                                                                 // 410
      start = applied.value;                                                                                 // 411
    }                                                                                                        // 412
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                           // 413
                                                                                                             // 414
    for (var i = start ; i < MIGRATIONS.length ; i++) {                                                      // 415
      // Apply migration i, then record that migration i was successfully run.                               // 416
      console.log("Applying migration " + (i+1));                                                            // 417
      MIGRATIONS[i]();                                                                                       // 418
      Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                                  // 419
      console.log("Applied migration " + (i+1));                                                             // 420
    }                                                                                                        // 421
  }                                                                                                          // 422
}                                                                                                            // 423
                                                                                                             // 424
SandstormDb.prototype.migrateToLatest = migrateToLatest;                                                     // 425
                                                                                                             // 426
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
